import os
import sys
import time

from monk.pip_unit_tests.gluon.test_optimizer_sgd import test_optimizer_sgd
from monk.pip_unit_tests.gluon.test_optimizer_nesterov_sgd import test_optimizer_nesterov_sgd
from monk.pip_unit_tests.gluon.test_optimizer_rmsprop import test_optimizer_rmsprop
from monk.pip_unit_tests.gluon.test_optimizer_momentum_rmsprop import test_optimizer_momentum_rmsprop
from monk.pip_unit_tests.gluon.test_optimizer_adam import test_optimizer_adam
from monk.pip_unit_tests.gluon.test_optimizer_adamax import test_optimizer_adamax
from monk.pip_unit_tests.gluon.test_optimizer_adadelta import test_optimizer_adadelta
from monk.pip_unit_tests.gluon.test_optimizer_adagrad import test_optimizer_adagrad
from monk.pip_unit_tests.gluon.test_optimizer_nadam import test_optimizer_nadam
from monk.pip_unit_tests.gluon.test_optimizer_signum import test_optimizer_signum

from monk.pip_unit_tests.gluon.test_loss_l1 import test_loss_l1
from monk.pip_unit_tests.gluon.test_loss_l2 import test_loss_l2
from monk.pip_unit_tests.gluon.test_loss_softmax_crossentropy import test_loss_softmax_crossentropy
from monk.pip_unit_tests.gluon.test_loss_crossentropy import test_loss_crossentropy
from monk.pip_unit_tests.gluon.test_loss_sigmoid_binary_crossentropy import test_loss_sigmoid_binary_crossentropy
from monk.pip_unit_tests.gluon.test_loss_binary_crossentropy import test_loss_binary_crossentropy
from monk.pip_unit_tests.gluon.test_loss_kldiv import test_loss_kldiv
from monk.pip_unit_tests.gluon.test_loss_poisson_nll import test_loss_poisson_nll
from monk.pip_unit_tests.gluon.test_loss_huber import test_loss_huber
from monk.pip_unit_tests.gluon.test_loss_hinge import test_loss_hinge
from monk.pip_unit_tests.gluon.test_loss_squared_hinge import test_loss_squared_hinge


from monk.pip_unit_tests.gluon.test_layer_convolution1d import test_layer_convolution1d
from monk.pip_unit_tests.gluon.test_layer_convolution2d import test_layer_convolution2d
from monk.pip_unit_tests.gluon.test_layer_convolution3d import test_layer_convolution3d
from monk.pip_unit_tests.gluon.test_layer_transposed_convolution1d import test_layer_transposed_convolution1d
from monk.pip_unit_tests.gluon.test_layer_transposed_convolution2d import test_layer_transposed_convolution2d
from monk.pip_unit_tests.gluon.test_layer_transposed_convolution3d import test_layer_transposed_convolution3d
from monk.pip_unit_tests.gluon.test_layer_max_pooling1d import test_layer_max_pooling1d
from monk.pip_unit_tests.gluon.test_layer_max_pooling2d import test_layer_max_pooling2d
from monk.pip_unit_tests.gluon.test_layer_max_pooling3d import test_layer_max_pooling3d
from monk.pip_unit_tests.gluon.test_layer_average_pooling1d import test_layer_average_pooling1d
from monk.pip_unit_tests.gluon.test_layer_average_pooling2d import test_layer_average_pooling2d
from monk.pip_unit_tests.gluon.test_layer_average_pooling3d import test_layer_average_pooling3d
from monk.pip_unit_tests.gluon.test_layer_global_max_pooling1d import test_layer_global_max_pooling1d
from monk.pip_unit_tests.gluon.test_layer_global_max_pooling2d import test_layer_global_max_pooling2d
from monk.pip_unit_tests.gluon.test_layer_global_max_pooling3d import test_layer_global_max_pooling3d
from monk.pip_unit_tests.gluon.test_layer_global_average_pooling1d import test_layer_global_average_pooling1d
from monk.pip_unit_tests.gluon.test_layer_global_average_pooling2d import test_layer_global_average_pooling2d
from monk.pip_unit_tests.gluon.test_layer_global_average_pooling3d import test_layer_global_average_pooling3d
from monk.pip_unit_tests.gluon.test_layer_batch_normalization import test_layer_batch_normalization
from monk.pip_unit_tests.gluon.test_layer_instance_normalization import test_layer_instance_normalization
from monk.pip_unit_tests.gluon.test_layer_layer_normalization import test_layer_layer_normalization
from monk.pip_unit_tests.gluon.test_layer_identity import test_layer_identity
from monk.pip_unit_tests.gluon.test_layer_fully_connected import test_layer_fully_connected
from monk.pip_unit_tests.gluon.test_layer_dropout import test_layer_dropout
from monk.pip_unit_tests.gluon.test_layer_flatten import test_layer_flatten
from monk.pip_unit_tests.gluon.test_activation_relu import test_activation_relu
from monk.pip_unit_tests.gluon.test_activation_sigmoid import test_activation_sigmoid 
from monk.pip_unit_tests.gluon.test_activation_tanh import test_activation_tanh
from monk.pip_unit_tests.gluon.test_activation_softplus import test_activation_softplus
from monk.pip_unit_tests.gluon.test_activation_softsign import test_activation_softsign
from monk.pip_unit_tests.gluon.test_activation_elu import test_activation_elu
from monk.pip_unit_tests.gluon.test_activation_gelu import test_activation_gelu
from monk.pip_unit_tests.gluon.test_activation_prelu import test_activation_prelu
from monk.pip_unit_tests.gluon.test_activation_leaky_relu import test_activation_leaky_relu
from monk.pip_unit_tests.gluon.test_activation_selu import test_activation_selu
from monk.pip_unit_tests.gluon.test_activation_swish import test_activation_swish
from monk.pip_unit_tests.gluon.test_layer_concatenate import test_layer_concatenate

from monk.pip_unit_tests.gluon.test_initializer_xavier_normal import test_initializer_xavier_normal
from monk.pip_unit_tests.gluon.test_initializer_xavier_uniform import test_initializer_xavier_uniform
from monk.pip_unit_tests.gluon.test_initializer_orthogonal_normal import test_initializer_orthogonal_normal
from monk.pip_unit_tests.gluon.test_initializer_orthogonal_uniform import test_initializer_orthogonal_uniform
from monk.pip_unit_tests.gluon.test_initializer_normal import test_initializer_normal
from monk.pip_unit_tests.gluon.test_initializer_uniform import test_initializer_uniform
from monk.pip_unit_tests.gluon.test_initializer_msra import test_initializer_msra

from monk.pip_unit_tests.gluon.test_block_resnet_v1 import test_block_resnet_v1
from monk.pip_unit_tests.gluon.test_block_resnet_v2 import test_block_resnet_v2
from monk.pip_unit_tests.gluon.test_block_resnet_v1_bottleneck import test_block_resnet_v1_bottleneck
from monk.pip_unit_tests.gluon.test_block_resnet_v2_bottleneck import test_block_resnet_v2_bottleneck
from monk.pip_unit_tests.gluon.test_block_resnext import test_block_resnext
from monk.pip_unit_tests.gluon.test_block_mobilenet_v2_linear_bottleneck import test_block_mobilenet_v2_linear_bottleneck
from monk.pip_unit_tests.gluon.test_block_mobilenet_v2_inverted_linear_bottleneck import test_block_mobilenet_v2_inverted_linear_bottleneck
from monk.pip_unit_tests.gluon.test_block_squeezenet_fire import test_block_squeezenet_fire
from monk.pip_unit_tests.gluon.test_block_densenet import test_block_densenet
from monk.pip_unit_tests.gluon.test_block_conv_bn_relu import test_block_conv_bn_relu
from monk.pip_unit_tests.gluon.test_block_inception_a import test_block_inception_a
from monk.pip_unit_tests.gluon.test_block_inception_b import test_block_inception_b
from monk.pip_unit_tests.gluon.test_block_inception_c import test_block_inception_c
from monk.pip_unit_tests.gluon.test_block_inception_d import test_block_inception_d
from monk.pip_unit_tests.gluon.test_block_inception_e import test_block_inception_e



from monk.pip_functionality_tests.gluon.test_default_train import test_default_train
from monk.pip_functionality_tests.gluon.test_default_eval_infer import test_default_eval_infer
from monk.pip_functionality_tests.gluon.test_update_copy_from import test_update_copy_from
from monk.pip_functionality_tests.gluon.test_update_normal import test_update_normal
from monk.pip_functionality_tests.gluon.test_update_eval_infer import test_update_eval_infer
from monk.pip_functionality_tests.gluon.test_expert_train import test_expert_train
from monk.pip_functionality_tests.gluon.test_expert_eval_infer import test_expert_eval_infer
from monk.pip_functionality_tests.gluon.test_switch_default import test_switch_default
from monk.pip_functionality_tests.gluon.test_switch_expert import test_switch_expert
from monk.pip_functionality_tests.gluon.test_compare import test_compare
from monk.pip_functionality_tests.gluon.test_analyse import test_analyse



def run_functionality_tests():
	origstdout = sys.stdout

	print("Running Tests...");
	sys.stdout = open("test_logs.txt", 'w');


	system_dict = {};
	system_dict["total_tests"] = 0;
	system_dict["successful_tests"] = 0;
	system_dict["failed_tests_lists"] = [];
	system_dict["failed_tests_exceptions"] = [];
	system_dict["skipped_tests_lists"] = [];



	start = time.time()


	print("Running 1/11");
	system_dict = test_default_train(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running 2/11");
	sys.stdout = open("test_logs.txt", 'a');
	system_dict = test_default_eval_infer(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));
	print("")


	print("Running 3/11");
	sys.stdout = open("test_logs.txt", 'a');
	system_dict = test_update_copy_from(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));
	print("")



	print("Running 4/11");
	sys.stdout = open("test_logs.txt", 'a');
	system_dict = test_update_normal(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));
	print("")



	print("Running 5/11");
	sys.stdout = open("test_logs.txt", 'a');
	system_dict = test_update_eval_infer(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));
	print("")



	print("Running 6/11");
	sys.stdout = open("test_logs.txt", 'a');
	system_dict = test_expert_train(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));



	print("Running 7/11");
	sys.stdout = open("test_logs.txt", 'a');
	system_dict = test_expert_eval_infer(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));
	print("")




	print("Running 8/11");
	sys.stdout = open("test_logs.txt", 'a');
	system_dict = test_switch_default(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));
	print("")




	print("Running 9/11");
	sys.stdout = open("test_logs.txt", 'a');
	system_dict = test_switch_expert(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));
	print("")




	print("Running 10/11");
	sys.stdout = open("test_logs.txt", 'a');
	system_dict = test_compare(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));
	print("")


	print("Running 11/11");
	sys.stdout = open("test_logs.txt", 'a');
	system_dict = test_analyse(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));
	print("")



	sys.stdout = open("test_logs.txt", 'a');
	end = time.time();

	print("Total Tests           - {}".format(system_dict["total_tests"]));
	print("Time Taken            - {} sec".format(end-start));
	print("Num Successful Tests  - {}".format(system_dict["successful_tests"]));
	print("Num Failed Tests      - {}".format(len(system_dict["failed_tests_lists"])));
	print("Num Skipped Tests     - {}".format(len(system_dict["skipped_tests_lists"])));
	print("");


	for i in range(len(system_dict["failed_tests_lists"])):
	    print("{}. Failed Test:".format(i+1));
	    print("Name     - {}".format(system_dict["failed_tests_lists"][i]));
	    print("Error    - {}".format(system_dict["failed_tests_exceptions"][i]));
	    print("");


	print("Skipped Tests List    - {}".format(system_dict["skipped_tests_lists"]));
	print("");


	sys.stdout = origstdout;


	print("Total Tests           - {}".format(system_dict["total_tests"]));
	print("Time Taken            - {} sec".format(end-start));
	print("Num Successful Tests  - {}".format(system_dict["successful_tests"]));
	print("Num Failed Tests      - {}".format(len(system_dict["failed_tests_lists"])));
	print("Num Skipped Tests     - {}".format(len(system_dict["skipped_tests_lists"])));
	print("See test_logs.txt for errors");
	print("");


	os.system("rm -r workspace");







def run_unit_tests():

	origstdout = sys.stdout


	print("Running Tests...");
	sys.stdout = open("test_logs.txt", 'w');


	system_dict = {};
	system_dict["total_tests"] = 0;
	system_dict["successful_tests"] = 0;
	system_dict["failed_tests_lists"] = [];
	system_dict["failed_tests_exceptions"] = [];
	system_dict["skipped_tests_lists"] = [];


	start = time.time()

	exp_num = 1;

	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_max_pooling1d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_max_pooling2d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_max_pooling3d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_average_pooling1d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_average_pooling2d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_average_pooling3d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_global_max_pooling1d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_global_max_pooling2d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_global_max_pooling3d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_global_average_pooling1d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_global_average_pooling2d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_global_average_pooling3d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")



	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_optimizer_sgd(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_optimizer_nesterov_sgd(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")

	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_optimizer_rmsprop(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_optimizer_momentum_rmsprop(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")

	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_optimizer_adam(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_optimizer_adamax(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_optimizer_adadelta(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_optimizer_adagrad(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")



	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_optimizer_nadam(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")



	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_optimizer_signum(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")







	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_loss_l1(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_loss_l2(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")



	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_loss_softmax_crossentropy(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")




	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_loss_crossentropy(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_loss_sigmoid_binary_crossentropy(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")



	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_loss_binary_crossentropy(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")



	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_loss_kldiv(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_loss_poisson_nll(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")



	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_loss_huber(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_loss_hinge(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")



	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_loss_squared_hinge(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")



















	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_batch_normalization(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_instance_normalization(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_layer_normalization(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_identity(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_fully_connected(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")



	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_dropout(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_flatten(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")





	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_activation_relu(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")



	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_activation_sigmoid(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_activation_tanh(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_activation_softplus(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_activation_softsign(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")



	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_activation_elu(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_activation_gelu(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_activation_prelu(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_activation_leaky_relu(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_activation_selu(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_activation_swish(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_concatenate(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")





	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_initializer_xavier_normal(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")

	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_initializer_xavier_uniform(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")

	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_initializer_orthogonal_normal(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")

	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_initializer_orthogonal_uniform(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")

	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_initializer_normal(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")

	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_initializer_uniform(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")

	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_initializer_msra(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")










	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_block_resnet_v1(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_block_resnet_v2(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_block_resnet_v1_bottleneck(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_block_resnet_v2_bottleneck(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_block_resnext(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_block_mobilenet_v2_linear_bottleneck(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")



	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_block_mobilenet_v2_inverted_linear_bottleneck(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_block_squeezenet_fire(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_block_densenet(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_block_conv_bn_relu(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_block_inception_a(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_block_inception_b(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_block_inception_c(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_block_inception_d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_block_inception_e(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")




	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_convolution1d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")



	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_convolution2d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_convolution3d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_transposed_convolution1d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_transposed_convolution2d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")


	print("Running {}/<num>".format(exp_num));
	exp_num += 1;
	system_dict = test_layer_transposed_convolution3d(system_dict)
	sys.stdout = origstdout;
	print("Tests Completed           - {}".format(system_dict["total_tests"]));
	print("Tests Succesful           - {}".format(system_dict["successful_tests"]));    
	print("")




	sys.stdout = open("test_logs.txt", 'a');
	end = time.time();

	print("Total Tests           - {}".format(system_dict["total_tests"]));
	print("Time Taken            - {} sec".format(end-start));
	print("Num Successful Tests  - {}".format(system_dict["successful_tests"]));
	print("Num Failed Tests      - {}".format(len(system_dict["failed_tests_lists"])));
	print("Num Skipped Tests     - {}".format(len(system_dict["skipped_tests_lists"])));
	print("");



	for i in range(len(system_dict["failed_tests_lists"])):
	    print("{}. Failed Test:".format(i+1));
	    print("Name     - {}".format(system_dict["failed_tests_lists"][i]));
	    print("Error    - {}".format(system_dict["failed_tests_exceptions"][i]));
	    print("");


	print("Skipped Tests List    - {}".format(system_dict["skipped_tests_lists"]));
	print("");


	sys.stdout = origstdout;


	print("Total Tests           - {}".format(system_dict["total_tests"]));
	print("Time Taken            - {} sec".format(end-start));
	print("Num Successful Tests  - {}".format(system_dict["successful_tests"]));
	print("Num Failed Tests      - {}".format(len(system_dict["failed_tests_lists"])));
	print("Num Skipped Tests     - {}".format(len(system_dict["skipped_tests_lists"])));
	print("See test_logs.txt for errors");
	print("");


	os.system("rm -r workspace");


from monk.pytorch.finetune.imports import *
from monk.system.imports import *
from monk.pytorch.finetune.level_1_dataset_base import finetune_dataset


class finetune_model(finetune_dataset):
    '''
    Base class for Model setup

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''
    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose);


    ###############################################################################################################################################
    @accepts("self", path=[bool, str, list], post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_model_final(self, path=False):
        '''
        Setup model based on set parameters

        Args:
            path (str): Dummy variable

        Returns:
            None
        '''
        self.custom_print("Model Details");
        if(self.system_dict["model"]["params"]["model_path"]):
            if(os.path.isfile(self.system_dict["model"]["params"]["model_path"])):
                self.custom_print("    Loading model - {}".format(self.system_dict["model"]["params"]["model_path"]));
                self.system_dict["local"]["model"] = load_model(self.system_dict, external_path=self.system_dict["model"]["params"]["model_path"]);
                self.system_dict = model_to_device(self.system_dict);
                self.custom_print("    Model loaded!");
                self.custom_print("");

                self.system_dict = get_num_layers(self.system_dict);

                self.system_dict["local"]["params_to_update"] = []
                layer_names = [];
                for name, param in self.system_dict["local"]["model"].named_parameters():
                    if param.requires_grad == True:
                        self.system_dict["local"]["params_to_update"].append(param);
                        lname = ".".join(name.split(".")[:-1]);
                        if lname not in layer_names:
                            layer_names.append(lname);
                self.system_dict["model"]["params"]["num_params_to_update"] = len(layer_names);
                self.system_dict["model"]["status"] = True;
                
            else:
                msg = "Model not found - {}\n".format(self.system_dict["model"]["params"]["model_path"]);
                msg += "Previous Training Incomplete.";
                raise ConstraintError(msg);

        
        elif(self.system_dict["states"]["eval_infer"]):
            if(os.path.isfile(self.system_dict["model_dir_relative"] + 'final')):
                self.custom_print("    Loading model - {}".format(self.system_dict["model_dir_relative"] + 'final'));
                self.system_dict["local"]["model"] = load_model(self.system_dict, final=True);
                self.system_dict = model_to_device(self.system_dict);
                self.custom_print("    Model loaded!");
                self.custom_print("");
            else:
                msg = "Model not found - {}\n".format(self.system_dict["model_dir_relative"] + 'final');
                msg += "Previous Training Incomplete.";
                raise ConstraintError(msg);
        else:
            if(self.system_dict["states"]["resume_train"]):
                if(os.path.isfile(self.system_dict["model_dir_relative"] + 'resume_state')):
                    self.custom_print("    Loading model - {}".format(self.system_dict["model_dir_relative"] + 'resume_state'));
                    self.system_dict["local"]["model"] = load_model(self.system_dict, resume=True);
                else:
                    msg = "Model not found - \"{}\"\n".format(self.system_dict["model_dir_relative"] + 'resume_state');
                    msg += "Training not started. Cannot Run resume Mode";
                    raise ConstraintError(msg);

            elif(self.system_dict["states"]["copy_from"]):
                model_path = self.system_dict["master_systems_dir_relative"] + self.system_dict["origin"][0] + "/" + self.system_dict["origin"][1] + "/output/models/";
                if(os.path.isfile(model_path + 'final')):
                    self.custom_print("    Loading model - {}".format(model_path + 'final'));
                    self.system_dict["local"]["model"] = load_model(self.system_dict, path=model_path, final=True);
                else:
                    msg = "Model not found - {}\n".format(model_path);
                    msg += "Previous Training Incomplete.";
                    raise ConstraintError(msg);

            else:
                self.custom_print("    Loading pretrained model");
                self.system_dict = setup_model(self.system_dict);


            
            self.system_dict = model_to_device(self.system_dict);
            self.custom_print("    Model Loaded on device");


            self.system_dict = get_num_layers(self.system_dict);

            self.system_dict["local"]["params_to_update"] = []
            layer_names = [];
            for name, param in self.system_dict["local"]["model"].named_parameters():
                if param.requires_grad == True:
                    self.system_dict["local"]["params_to_update"].append(param);
                    lname = ".".join(name.split(".")[:-1]);
                    if lname not in layer_names:
                        layer_names.append(lname);
            self.system_dict["model"]["params"]["num_params_to_update"] = len(layer_names);
            self.system_dict["model"]["status"] = True;

            
            if(self.system_dict["model"]["type"] == "custom"):
                self.custom_print("        Model name:                           {}".format("Custom Model"));
            else:
                self.custom_print("        Model name:                           {}".format(self.system_dict["model"]["params"]["model_name"]));
            self.custom_print("        Num layers in model:  {}".format(self.system_dict["model"]["params"]["num_layers"]));
            self.custom_print("        Num trainable layers: {}".format(self.system_dict["model"]["params"]["num_params_to_update"]));
            self.custom_print("");
            
    ###############################################################################################################################################


from monk.system.imports import *


@accepts(post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=True)
def get_base_system_dict():
    system_dict = {};

    system_dict["verbose"] = 1;
    system_dict["cwd"] = False;
    system_dict["master_systems_dir"] = False
    system_dict["master_systems_dir_relative"] = False;


    system_dict["project_name"] = False;
    system_dict["project_dir"] = False;
    system_dict["project_dir_relative"] = False;
    system_dict["experiment_name"] = False;
    system_dict["experiment_dir"] = False;
    system_dict["experiment_dir_relative"] = False;
    system_dict["origin"] = False;

    system_dict["master_comparison_dir"] = False;
    system_dict["master_comparison_dir_relative"] = False;
    system_dict["library"] = False;
    system_dict["output_dir"] = False;
    system_dict["output_dir_relative"] = False;
    system_dict["model_dir"] = False;
    system_dict["model_dir_relative"] = False;
    system_dict["log_dir"] = False;
    system_dict["log_dir_relative"] = False;
    system_dict["fname"] = False;
    system_dict["fname_relative"] = False;


    #Dataset details
    system_dict["dataset"] = {};
    system_dict["dataset"]["dataset_type"] = False
    system_dict["dataset"]["label_type"] = False;
    system_dict["dataset"]["train_path"] = False;
    system_dict["dataset"]["val_path"] = False;
    system_dict["dataset"]["csv_train"] = False;
    system_dict["dataset"]["csv_val"] = False;
    system_dict["dataset"]["test_path"] = False;
    system_dict["dataset"]["csv_test"] = False;


    #Dataset params
    system_dict["dataset"]["params"] = {};
    system_dict["dataset"]["params"]["input_size"] = False;
    system_dict["dataset"]["params"]["data_shape"] = False;
    system_dict["dataset"]["params"]["batch_size"] = False;
    system_dict["dataset"]["params"]["train_shuffle"] = False;
    system_dict["dataset"]["params"]["train_shuffle"] = False;
    system_dict["dataset"]["params"]["num_workers"] = False;
    system_dict["dataset"]["params"]["weighted_sample"] = False;
    system_dict["dataset"]["params"]["num_classes"] = False;
    system_dict["dataset"]["params"]["classes"] = False;
    system_dict["dataset"]["params"]["num_train_images"] = False;
    system_dict["dataset"]["params"]["num_val_images"] = False;
    system_dict["dataset"]["params"]["num_test_images"] = False;
    system_dict["dataset"]["params"]["delimiter"] = ",";
    system_dict["dataset"]["params"]["test_delimiter"] = ",";
    system_dict["dataset"]["params"]["dataset_test_type"] = False;
    system_dict["dataset"]["params"]["train_val_split"] = 0.9;


    #Dataset transforms
    system_dict["dataset"]["transforms"] = {};
    system_dict["dataset"]["transforms"]["train"] = [];
    system_dict["dataset"]["transforms"]["val"] = [];
    system_dict["dataset"]["transforms"]["test"] = [];
    system_dict["dataset"]["status"] = False;

    #Model details
    system_dict["model"] = {};
    system_dict["model"]["status"] = False;
    system_dict["model"]["final_layer"] = False;
    system_dict["model"]["type"] = "pretrained";
    system_dict["model"]["custom_network"] = [];

    #Custom Model details
    system_dict["custom_model"] = {};
    system_dict["custom_model"]["status"] = False;
    system_dict["custom_model"]["network_stack"] = [];
    system_dict["custom_model"]["network_initializer"] = False;

    
    #Model params
    system_dict["model"]["params"] = {};
    system_dict["model"]["params"]["model_name"] = False;
    system_dict["model"]["params"]["model_path"] = False;
    system_dict["model"]["params"]["use_gpu"] = False;
    system_dict["model"]["params"]["use_pretrained"] = False;
    system_dict["model"]["params"]["freeze_base_network"] = False;
    system_dict["model"]["params"]["num_layers"] = False;
    system_dict["model"]["params"]["num_params_to_update"] = False;
    system_dict["model"]["params"]["num_freeze"] = False;
    system_dict["model"]["params"]["gpu_memory_fraction"] = 0.6;



    #Hyper parameter details
    system_dict["hyper-parameters"] = {};
    system_dict["hyper-parameters"]["status"] = False;
    system_dict["hyper-parameters"]["learning_rate"] = False;
    system_dict["hyper-parameters"]["num_epochs"] = False;
    system_dict["hyper-parameters"]["optimizer"] = {};
    system_dict["hyper-parameters"]["optimizer"]["name"] = False;
    system_dict["hyper-parameters"]["optimizer"]["params"] = {};
    system_dict["hyper-parameters"]["learning_rate_scheduler"] = {};
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"] = False;
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"] = {};
    system_dict["hyper-parameters"]["loss"] = {};
    system_dict["hyper-parameters"]["loss"]["name"] = False;
    system_dict["hyper-parameters"]["loss"]["params"] = {};

    #Training details
    system_dict["training"] = {};
    system_dict["training"]["settings"] = {};
    system_dict["training"]["settings"]["display_progress_realtime"] = False;
    system_dict["training"]["settings"]["display_progress"] = False;
    system_dict["training"]["settings"]["save_intermediate_models"] = False;
    system_dict["training"]["settings"]["save_training_logs"] = False;
    system_dict["training"]["settings"]["intermediate_model_prefix"] = False;
    system_dict["training"]["outputs"] = {};
    system_dict["training"]["outputs"]["max_gpu_memory_usage"] = 0;
    system_dict["training"]["outputs"]["best_val_acc"] = 0;
    system_dict["training"]["outputs"]["best_val_acc_epoch_num"] = 0;
    system_dict["training"]["outputs"]["epochs_completed"] = 0;
    system_dict["training"]["status"] = False;

    #Testing details
    system_dict["testing"] = {};
    system_dict["testing"]["status"] = False;
    system_dict["testing"]["num_images"] = False;
    system_dict["testing"]["num_correct_predictions"] = False;
    system_dict["testing"]["percentage_accuracy"] = False;
    system_dict["testing"]["class_accuracy"] = False;

    #States
    system_dict["states"] = {};
    system_dict["states"]["eval_infer"] = False;
    system_dict["states"]["resume_train"] = False;
    system_dict["states"]["copy_from"] = False;
    system_dict["states"]["pseudo_copy_from"] = False;

    #Local variables
    system_dict["local"] = {};

    system_dict["local"]["projects_list"] = [];
    system_dict["local"]["num_projects"] = False;
    system_dict["local"]["experiments_list"] = [];
    system_dict["local"]["num_experiments"] = False;
    system_dict["local"]["project_experiment_list"] = [];

    system_dict["local"]["transforms_train"] = [];
    system_dict["local"]["transforms_val"] = [];
    system_dict["local"]["transforms_test"] = [];
    system_dict["local"]["normalize"] = False;
    system_dict["local"]["mean_subtract"] = False;
    system_dict["local"]["applied_train_tensor"] = False;
    system_dict["local"]["applied_test_tensor"] = False;
    system_dict["local"]["data_transforms"] = {};
    system_dict["local"]["image_datasets"] = {};
    system_dict["local"]["data_loaders"] = {};
    system_dict["local"]["data_generators"] = {};

    system_dict["local"]["model"] = False;
    system_dict["local"]["custom_model"] = False;
    system_dict["local"]["ctx"] = False;
    system_dict["local"]["params_to_update"] = [];
    system_dict["local"]["device"] = False;

    system_dict["local"]["learning_rate_scheduler"] = False;
    system_dict["local"]["optimizer"] = False;
    system_dict["local"]["criterion"] = False;

    #CNN Visualization storage details
    system_dict["visualization"] = {}
    
    system_dict["visualization"]["base"] = False;
    system_dict["visualization"]["kernels_dir"] = False;
    system_dict["visualization"]["feature_maps_dir"] = False;

    return system_dict;

#@accepts(post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def update_local_var(system_dict):
    system_dict["local"] = {};

    system_dict["local"]["projects_list"] = [];
    system_dict["local"]["num_projects"] = False;
    system_dict["local"]["experiments_list"] = [];
    system_dict["local"]["num_experiments"] = False;
    system_dict["local"]["project_experiment_list"] = [];

    system_dict["local"]["transforms_train"] = [];
    system_dict["local"]["transforms_val"] = [];
    system_dict["local"]["transforms_test"] = [];
    system_dict["local"]["normalize"] = False;
    system_dict["local"]["mean_subtract"] = False;
    system_dict["local"]["applied_train_tensor"] = False;
    system_dict["local"]["applied_test_tensor"] = False;
    system_dict["local"]["data_transforms"] = {};
    system_dict["local"]["image_datasets"] = {};
    system_dict["local"]["data_loaders"] = {};
    system_dict["local"]["data_generators"] = {};

    system_dict["local"]["model"] = False;
    system_dict["local"]["custom_model"] = False;
    system_dict["local"]["ctx"] = False;
    system_dict["local"]["params_to_update"] = [];
    system_dict["local"]["device"] = False;

    system_dict["local"]["learning_rate_scheduler"] = False;
    system_dict["local"]["optimizer"] = False;
    system_dict["local"]["criterion"] = False;


    system_dict["states"] = {};
    system_dict["states"]["eval_infer"] = False;
    system_dict["states"]["resume_train"] = False;
    system_dict["states"]["copy_from"] = False;
    system_dict["states"]["pseudo_copy_from"] = False;

    return system_dict;


import os
import sys

import psutil

from monk.pytorch_prototype import prototype
from monk.compare_prototype import compare
from monk.pip_functionality_tests.pytorch.common import print_start
from monk.pip_functionality_tests.pytorch.common import print_status



def test_update_copy_from(system_dict):
    forward = True;
    if(not os.path.isdir("datasets")):
        os.system("! wget --load-cookies /tmp/cookies.txt \"https://docs.google.com/uc?export=download&confirm=$(wget --save-cookies /tmp/cookies.txt --keep-session-cookies --no-check-certificate 'https://docs.google.com/uc?export=download&id=1rG-U1mS8hDU7_wM56a1kc-li_zHLtbq2' -O- | sed -rn 's/.*confirm=([0-9A-Za-z_]+).*/\1\n/p')&id=1rG-U1mS8hDU7_wM56a1kc-li_zHLtbq2\" -O datasets.zip && rm -rf /tmp/cookies.txt")
        os.system("! unzip -qq datasets.zip")

    test = "update_copy_from_object_creation";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ptf = prototype(verbose=0);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "update_copy_from_Prototype()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ptf.Prototype("sample-project-1", "sample-experiment-2", copy_from=["sample-project-1", "sample-experiment-1"]);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "update_copy_from_reset_transforms()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ptf.reset_transforms();
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");




    test = "update_copy_from_apply_transforms()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ptf.apply_random_resized_crop(256, train=True, val=True, test=True);
            ptf.apply_random_perspective(train=True, val=True);
            ptf.apply_random_vertical_flip(train=True, val=True);
            ptf.apply_random_horizontal_flip(train=True, val=True);
            ptf.apply_normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225], train=True, val=True, test=True);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "update_copy_from_update_dataset()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ptf.update_dataset(dataset_path=["datasets/dataset_cats_dogs_train", "datasets/dataset_cats_dogs_eval"]);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");



    test = "update_copy_from_update_input_size()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ptf.update_input_size(256);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "update_copy_from_update_batch_size()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ptf.update_batch_size(6);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "update_copy_from_update_shuffle_data()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ptf.update_shuffle_data(False);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "update_copy_from_update_num_processors()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ptf.update_num_processors(16);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "update_copy_from_update_trainval_split()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ptf.update_trainval_split(0.6);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "update_copy_from_Reload()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ptf.Reload();
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "update_copy_from_EDA()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ptf.EDA(check_missing=True, check_corrupt=True);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "update_copy_from_Train()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ptf.Train();
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");








    return system_dict


from monk.tf_keras_1.finetune.imports import *
from monk.system.imports import *

from monk.tf_keras_1.finetune.level_8_layers_main import prototype_layers


class prototype_transforms(prototype_layers):
    '''
    Main class for all transforms in expert mode

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''
    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose);
        tmp = {};
        tmp["featurewise_center"] = False;
        tmp["featurewise_std_normalization"] = False;
        tmp["rotation_range"] = 0;
        tmp["width_shift_range"] = 0;
        tmp["height_shift_range"] = 0;
        tmp["shear_range"] = 0;
        tmp["zoom_range"] = 0;
        tmp["brightness_range"] = None;
        tmp["horizontal_flip"] = False;
        tmp["vertical_flip"] = False;
        tmp["mean"] = False;
        tmp["std"] = False;

        self.system_dict["local"]["transforms_train"] = tmp;
        self.system_dict["local"]["transforms_val"] = tmp;
        self.system_dict["local"]["transforms_test"] = tmp;





    ###############################################################################################################################################
    @error_checks(None, brightness=["gte", 0.0], contrast=["gte", 0.0], saturation=["gte", 0.0], hue=["gte", 0.0], 
        train=None, val=None, test=None, post_trace=False)
    @accepts("self", brightness=[int, float], contrast=[int, float], saturation=[int, float], hue=[int, float], 
        train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_color_jitter(self, brightness=0, contrast=0, saturation=0, hue=0, train=False, val=False, test=False):
        '''
        Apply Color jittering transformations

        Args:
            brightness (float): Levels to jitter brightness.
                                        0 - min
                                        1 - max
            contrast (float): Levels to jitter contrast.
                                        0 - min
                                        1 - max
            saturation (float): Levels to jitter saturation.
                                        0 - min
                                        1 - max
            hue (float): Levels to jitter hue.
                                        0 - min
                                        1 - max
            train (bool): If True, transform applied to training data
            val (bool): If True, transform applied to validation data
            test (bool): If True, transform applied to testing/inferencing data

        Returns:
            None
        '''
        self.system_dict = transform_color_jitter(self.system_dict, brightness, contrast, saturation, hue, train, val, test);
    ###############################################################################################################################################




    ###############################################################################################################################################
    @warning_checks(None, ["lt", 30], translate=["lt", 1.0], scale=["lt", 3.0], shear=None, 
        train=None, val=None, test=None, post_trace=False)
    @error_checks(None, ["gte", 0.0], translate=["gte", 0.0], scale=["gte", 0.0], shear=["gte", 0.0], 
        train=None, val=None, test=None, post_trace=False)
    @accepts("self", [list, float, int], translate=[tuple, type(None)], scale=[tuple, type(None)], 
        shear=[list, float, int, tuple, type(None)], train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_random_affine(self, degrees, translate=None, scale=None, shear=None, train=False, val=False, test=False):
        '''
        Apply random affine transformations

        Args:
            degrees (float): Max Rotation range limit for transforms
            scale (float, list): Range for randomly scaling 
            shear (float, list): Range for randomly applying sheer changes
            train (bool): If True, transform applied to training data
            val (bool): If True, transform applied to validation data
            test (bool): If True, transform applied to testing/inferencing data

        Returns:
            None
        '''
        self.system_dict = transform_random_affine(self.system_dict, degrees, translate, scale, shear, train, val, test);
    ###############################################################################################################################################



    ###############################################################################################################################################
    @error_checks(None, probability=["gt", 0, "lt", 1], train=None, val=None, test=None, post_trace=False)
    @accepts("self", probability=float, train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_random_horizontal_flip(self, probability=0.5, train=False, val=False, test=False):
        '''
        Apply random horizontal flip transformations

        Args:
            probability (float): Probability of flipping the input image
            train (bool): If True, transform applied to training data
            val (bool): If True, transform applied to validation data
            test (bool): If True, transform applied to testing/inferencing data

        Returns:
            None
        '''
        self.system_dict = transform_random_horizontal_flip(self.system_dict, probability, train, val, test);
    ###############################################################################################################################################



    ###############################################################################################################################################
    @error_checks(None, probability=["gt", 0, "lt", 1], train=None, val=None, test=None, post_trace=False)
    @accepts("self", probability=float, train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_random_vertical_flip(self, probability=0.5, train=False, val=False, test=False):
        '''
        Apply random vertical flip transformations

        Args:
            probability (float): Probability of flipping the input image
            train (bool): If True, transform applied to training data
            val (bool): If True, transform applied to validation data
            test (bool): If True, transform applied to testing/inferencing data

        Returns:
            None
        '''
        self.system_dict = transform_random_vertical_flip(self.system_dict, probability, train, val, test);
    ###############################################################################################################################################



    ###############################################################################################################################################
    @warning_checks(None, ["lt", 30], train=None, val=None, test=None, post_trace=False)
    @error_checks(None, ["gte", 0.0], train=None, val=None, test=None, post_trace=False)
    @accepts("self", [float, int, list], train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_random_rotation(self, degrees, train=False, val=False, test=False):
        '''
        Apply random rotation transformations

        Args:
            degrees (float): Max Rotation range limit for transforms
            train (bool): If True, transform applied to training data
            val (bool): If True, transform applied to validation data
            test (bool): If True, transform applied to testing/inferencing data

        Returns:
            None
        '''
        self.system_dict = transform_random_rotation(self.system_dict, degrees, train, val, test);
    ###############################################################################################################################################




    ###############################################################################################################################################
    @error_checks(None, mean=["gt", 0, "lt", 1], train=None, val=None, test=None, post_trace=False)
    @accepts("self", mean=[list, float], train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_mean_subtraction(self, mean=[123, 116, 103], train=False, val=False, test=False):
        '''
        Apply mean subtraction

        Args:
            mean (float, list): Mean value for subtraction
            train (bool): If True, transform applied to training data
            val (bool): If True, transform applied to validation data
            test (bool): If True, transform applied to testing/inferencing data

        Returns:
            None
        '''
        self.system_dict = transform_mean_subtraction(self.system_dict, mean, train, val, test);
    ###############################################################################################################################################





    ###############################################################################################################################################
    @error_checks(None, mean=["gt", 0, "lt", 1], std=["gt", 0, "lt", 1], train=None, val=None, test=None, post_trace=False)
    @accepts("self", mean=[list, float], std=[list, float], train=bool, val=bool, test=bool, post_trace=False)
    #@TraceFunction(trace_args=False, trace_rv=False)
    def apply_normalize(self, mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225], train=False, val=False, test=False):
        '''
        Apply mean subtraction and standard normalization

        Args:
            mean (float, list): Mean value for subtraction
            std (float, list): Normalization factor
            train (bool): If True, transform applied to training data
            val (bool): If True, transform applied to validation data
            test (bool): If True, transform applied to testing/inferencing data

        Returns:
            None
        '''
        self.system_dict = transform_normalize(self.system_dict, mean, std, train, val, test);
    ###############################################################################################################################################



from monk.gluon.finetune.imports import *
from monk.system.imports import *

from monk.gluon.finetune.level_11_optimizers_main import prototype_optimizers


class prototype_losses(prototype_optimizers):
    '''
    Main class for all parameters in expert mode

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''
    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose);


    ###############################################################################################################################################
    @accepts("self", weight=[list, type(np.array([1, 2, 3])), float, type(None)], batch_axis=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def loss_l1(self, weight=None, batch_axis=0):
        '''
        Select L1 Loss

        Args:
            weight (float): global scalar for weight loss
            batch_axis (int): Axis representing number of elements in the batch - N

        Returns:
            None
        '''
        self.system_dict = l1(self.system_dict, weight=weight, batch_axis=batch_axis);

        self.custom_print("Loss");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["loss"]["name"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["loss"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################



    ###############################################################################################################################################
    @accepts("self", weight=[list, type(np.array([1, 2, 3])), float, type(None)], batch_axis=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def loss_l2(self, weight=1.0, batch_axis=0):
        '''
        Select L2 Loss

        Args:
            weight (float): global scalar for weight loss
            batch_axis (int): Axis representing number of elements in the batch - N

        Returns:
            None
        '''
        self.system_dict = l2(self.system_dict, weight=weight, batch_axis=batch_axis);

        self.custom_print("Loss");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["loss"]["name"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["loss"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################



    ###############################################################################################################################################
    @accepts("self", weight=[list, type(np.array([1, 2, 3])), float, type(None)], batch_axis=int,
        axis_to_sum_over=int, label_as_categories=bool, label_smoothing=bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def loss_softmax_crossentropy(self, weight=None, batch_axis=0, axis_to_sum_over=-1, 
                                    label_as_categories=True, label_smoothing=False):
        '''
        Select soaftmax crossentropy Loss - Auto softmax before applying loss 

        Args:
            weight (float): global scalar for weight loss
            batch_axis (int): Axis representing number of elements in the batch - N
            axis_to_sum_over (int): Set as -1
            label_as_categories (bool): Fixed as True
            label_smoothing (bool): If True, label smoothning is applied.

        Returns:
            None
        '''
        self.system_dict = softmax_crossentropy(self.system_dict, weight=weight, batch_axis=batch_axis,
                                                axis_to_sum_over=axis_to_sum_over, label_as_categories=label_as_categories, 
                                                label_smoothing=label_smoothing);

        self.custom_print("Loss");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["loss"]["name"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["loss"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################


    ###############################################################################################################################################
    @accepts("self", weight=[list, type(np.array([1, 2, 3])), float, type(None)], batch_axis=int,
        axis_to_sum_over=int, label_as_categories=bool, label_smoothing=bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def loss_crossentropy(self, weight=None, batch_axis=0, axis_to_sum_over=-1, 
                                    label_as_categories=True, label_smoothing=False):
        '''
        Select crossentropy Loss - Need to manually apply softmax

        Args:
            weight (float): global scalar for weight loss
            batch_axis (int): Axis representing number of elements in the batch - N
            axis_to_sum_over (int): Set as -1
            label_as_categories (bool): Fixed as True
            label_smoothing (bool): If True, label smoothning is applied.

        Returns:
            None
        '''
        self.system_dict = crossentropy(self.system_dict, weight=weight, batch_axis=batch_axis,
                                                axis_to_sum_over=axis_to_sum_over, label_as_categories=label_as_categories, 
                                                label_smoothing=label_smoothing);

        self.custom_print("Loss");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["loss"]["name"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["loss"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################



    ###############################################################################################################################################
    @accepts("self", weight=[list, type(np.array([1, 2, 3])), float, type(None)], batch_axis=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def loss_sigmoid_binary_crossentropy(self, weight=None, batch_axis=0):
        '''
        Select sigmoid binary crossentropy Loss - Auto sigmoid before applying loss 

        Args:
            weight (float): global scalar for weight loss
            batch_axis (int): Axis representing number of elements in the batch - N

        Returns:
            None
        '''
        self.system_dict = sigmoid_binary_crossentropy(self.system_dict, weight=weight, batch_axis=batch_axis);

        self.custom_print("Loss");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["loss"]["name"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["loss"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################



    ###############################################################################################################################################
    @accepts("self", weight=[list, type(np.array([1, 2, 3])), float, type(None)], batch_axis=int, post_trace=False)    
    #@TraceFunction(trace_args=True, trace_rv=True)
    def loss_binary_crossentropy(self, weight=None, batch_axis=0):
        '''
        Select binary crossentropy Loss - Need to manually apply sigmoid

        Args:
            weight (float): global scalar for weight loss
            batch_axis (int): Axis representing number of elements in the batch - N

        Returns:
            None
        '''
        self.system_dict = binary_crossentropy(self.system_dict, weight=weight, batch_axis=batch_axis);

        self.custom_print("Loss");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["loss"]["name"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["loss"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################


    ###############################################################################################################################################
    @accepts("self", log_pre_applied=bool, weight=[list, type(np.array([1, 2, 3])), float, type(None)], 
        batch_axis=int, axis_to_sum_over=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def loss_kldiv(self, log_pre_applied=False, weight=None, batch_axis=0, axis_to_sum_over=-1):
        '''
        Select lkdiv Loss

        Args:
            weight (float): global scalar for weight loss
            batch_axis (int): Axis representing number of elements in the batch - N
            axis_to_sum_over (int): Set as -1
            log_pre_applied (bool): If set as False, then logarithmic function is auto applied over target variables

        Returns:
            None
        '''
        self.system_dict = kldiv(self.system_dict, weight=weight, batch_axis=batch_axis,
                                axis_to_sum_over=axis_to_sum_over, log_pre_applied=log_pre_applied);

        self.custom_print("Loss");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["loss"]["name"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["loss"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################


    ###############################################################################################################################################
    @accepts("self", log_pre_applied=bool, weight=[list, type(np.array([1, 2, 3])), float, type(None)], 
        batch_axis=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def loss_poisson_nll(self, log_pre_applied=False, weight=None, batch_axis=0):
        '''
        Select poisson_nll Loss

        Args:
            weight (float): global scalar for weight loss
            batch_axis (int): Axis representing number of elements in the batch - N
            log_pre_applied (bool): If set as False, then logarithmic function is auto applied over target variables

        Returns:
            None
        '''
        self.system_dict = poisson_nll(self.system_dict, log_pre_applied=log_pre_applied,
                                        weight=weight, batch_axis=batch_axis);

        self.custom_print("Loss");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["loss"]["name"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["loss"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################


    ###############################################################################################################################################
    @accepts("self", weight=[list, type(np.array([1, 2, 3])), float, type(None)], batch_axis=int,
        threshold_for_mean_estimator=[int, float], post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def loss_huber(self, weight=None, batch_axis=0, threshold_for_mean_estimator=1):
        '''
        Select huber Loss

        Args:
            weight (float): global scalar for weight loss
            batch_axis (int): Axis representing number of elements in the batch - N
            threshold_for_mean_estimator (int): Threshold for trimmed mean estimator.

        Returns:
            None
        '''
        self.system_dict = huber(self.system_dict, threshold_for_mean_estimator=threshold_for_mean_estimator,
                                weight=weight, batch_axis=batch_axis);

        self.custom_print("Loss");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["loss"]["name"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["loss"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################


    ###############################################################################################################################################
    @accepts("self", weight=[list, type(np.array([1, 2, 3])), float, type(None)], batch_axis=int,
        margin=[int, float], post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def loss_hinge(self, weight=None, batch_axis=0, margin=1):
        '''
        Select hinge Loss

        Args:
            weight (float): global scalar for weight loss
            batch_axis (int): Axis representing number of elements in the batch - N
            margin (float): MArgin value.

        Returns:
            None
        '''
        self.system_dict = hinge(self.system_dict, margin=margin,
                                weight=weight, batch_axis=batch_axis);

        self.custom_print("Loss");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["loss"]["name"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["loss"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################



    ###############################################################################################################################################
    @accepts("self", weight=[list, type(np.array([1, 2, 3])), float, type(None)], batch_axis=int,
        margin=[int, float], post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def loss_squared_hinge(self, weight=None, batch_axis=0, margin=1):
        '''
        Select squared hinge Loss

        Args:
            weight (float): global scalar for weight loss
            batch_axis (int): Axis representing number of elements in the batch - N
            margin (float): MArgin value.

        Returns:
            None
        '''
        self.system_dict = squared_hinge(self.system_dict, margin=margin,
                                weight=weight, batch_axis=batch_axis);

        self.custom_print("Loss");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["loss"]["name"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["loss"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################


from __future__ import print_function
from __future__ import division

import warnings
warnings.filterwarnings("ignore", category=FutureWarning)
warnings.simplefilter(action='ignore', category=Warning)

import warnings
import torch
import GPUtil
import networkx as nx
import torchvision
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from matplotlib import pyplot as plt

import os
import ipywidgets as widgets
from ipywidgets import interact, fixed
from IPython.display import display, clear_output
import matplotlib.gridspec as gridspec
from torchvision import transforms

layer_names = ["convolution1d", "convolution2d", "convolution", "convolution3d", "transposed_convolution1d",
                    "transposed_convolution", "transposed_convolution2d", "transposed_convolution3d",
                    "max_pooling1d", "max_pooling2d", "max_pooling", "max_pooling3d", "average_pooling1d",
                    "average_pooling2d", "average_pooling", "average_pooling3d", "global_max_pooling1d",
                    "global_max_pooling2d", "global_max_pooling", "global_max_pooling3d", "global_average_pooling1d",
                    "global_average_pooling2d", "global_average_pooling", "global_average_pooling3d", "fully_connected", 
                    "flatten", "dropout", "identity", "batch_normalization", "instance_normalization", "layer_normalization",
                    "relu", "sigmoid", "tanh", "softplus", "softsign",  "elu", "leaky_relu", "prelu", "selu",
                    "hardshrink", "hardtanh", "logsigmoid", "relu6", "rrelu", "celu", "softshrink", "tanhshrink",
                    "threshold", "softmin", "softmax", "logsoftmax", "add", "concatenate"]


names = ["conv1d_", "conv2d_", "conv_", "conv3d_", "tconv1d_", "tconv_", "tconv2d_", "tconv3d_",
			"max-pool1d_", "max-pool2d_", "max-pool_", "max-pool3d_", "avg-pooling1d_",
            "avg-pool2d_", "avg-pool_", "avg-pool3d_", "global-max-pool1d_", "global-max-pool2d_", 
            "global-max-pool_", "global-max-pool3d_", "global-avg-pool1d_", "global-avg-pool2d_", "global-avg-pool_", 
            "global-avg-pool3d_", "fc_", "flatten_", "dropout_", "identity_", "bn_", "inorm_",
            "lnorm_", "relu_", "sigmoid_", "tanh_", "softplus_", "softsign_", "elu_", "leaky_relu_", 
            "prelu_", "selu_", "hardshrink_", "hardtanh_", "logsigmoid_", "relu6_", "rrelu_", "celu_", 
            "softshrink_", "tanhshrink_", "threshold_", "softmin_", "softmax_", "logsoftmax_",
            "add_", "concatenate_"];


from monk.pytorch.finetune.imports import *
from monk.system.imports import *

from monk.pytorch.finetune.level_3_training_base import finetune_training


class finetune_evaluation(finetune_training):
    '''
    Bae class for external validation and inferencing

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''
    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose);


    ###############################################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_evaluation_final(self):
        '''
        Main function for external validation post training

        Args:
            None

        Returns:
            float: Accuracy in percentage
            dict: Class based accuracy in percentage
        '''
        self.custom_print("Testing");
        self.system_dict["testing"]["status"] = False;
        if(self.system_dict["training"]["settings"]["display_progress_realtime"] and self.system_dict["verbose"]):
            pbar=tqdm(total=len(self.system_dict["local"]["data_loaders"]["test"]));

        running_corrects = 0
        class_dict = {};
        for i in range(len(self.system_dict["dataset"]["params"]["classes"])):
            class_dict[self.system_dict["dataset"]["params"]["classes"][i]] = {};
            class_dict[self.system_dict["dataset"]["params"]["classes"][i]]["num_images"] = 0;
            class_dict[self.system_dict["dataset"]["params"]["classes"][i]]["num_correct"] = 0;

        for inputs, labels in self.system_dict["local"]["data_loaders"]["test"]:
            if(self.system_dict["training"]["settings"]["display_progress_realtime"] and self.system_dict["verbose"]):
                pbar.update();

            inputs = inputs.to(self.system_dict["local"]["device"]);
            labels = labels.to(self.system_dict["local"]["device"]);

            outputs = self.system_dict["local"]["model"](inputs)
            _, preds = torch.max(outputs, 1)

            label = int(labels.data.cpu().numpy())
            pred = int(preds.data.cpu().numpy())
            class_dict[self.system_dict["dataset"]["params"]["classes"][label]]["num_images"] += 1;
            if(label == pred):
                class_dict[self.system_dict["dataset"]["params"]["classes"][label]]["num_correct"] += 1;
            running_corrects += torch.sum(preds == labels.data)

        accuracy = running_corrects.double() / len(self.system_dict["local"]["data_loaders"]["test"].dataset)

        self.custom_print("");
        self.custom_print("    Result");
        self.custom_print("        class based accuracies");
        for i in range(len(self.system_dict["dataset"]["params"]["classes"])):
            self.custom_print("            {}. {} - {} %".format(i, self.system_dict["dataset"]["params"]["classes"][i], 
                class_dict[self.system_dict["dataset"]["params"]["classes"][i]]["num_correct"]/class_dict[self.system_dict["dataset"]["params"]["classes"][i]]["num_images"]*100));
            class_dict[self.system_dict["dataset"]["params"]["classes"][i]]["accuracy(%)"] = class_dict[self.system_dict["dataset"]["params"]["classes"][i]]["num_correct"]/class_dict[self.system_dict["dataset"]["params"]["classes"][i]]["num_images"]*100;
        self.custom_print("        total images:            {}".format(len(self.system_dict["local"]["data_loaders"]["test"])));
        self.custom_print("        num correct predictions: {}".format(int(running_corrects.cpu().numpy())));
        self.custom_print("        Average accuracy (%):    {}".format(accuracy.cpu().numpy()*100));

        self.system_dict["testing"]["num_images"] = len(self.system_dict["local"]["data_loaders"]["test"]);
        self.system_dict["testing"]["num_correct_predictions"] = int(running_corrects.cpu().numpy());
        self.system_dict["testing"]["percentage_accuracy"] = accuracy.cpu().numpy()*100;
        self.system_dict["testing"]["class_accuracy"] = class_dict
        self.system_dict["testing"]["status"] = True;
        self.custom_print("");
        return accuracy.cpu().numpy()*100, class_dict;
    ###############################################################################################################################################



    ###############################################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_evaluation_final_multiple(self):
        '''
        Main function for external validation post training for multi-label data

        Args:
            None

        Returns:
            float: Accuracy in percentage
            dict: Class based accuracy in percentage
        '''

        self.custom_print("Testing");
        self.system_dict["testing"]["status"] = False;
        if(self.system_dict["training"]["settings"]["display_progress_realtime"] and self.system_dict["verbose"]):
            pbar=tqdm(total=len(self.system_dict["local"]["data_loaders"]["test"]));

        running_corrects = 0
        total_labels = 0
        class_dict = {};

        for i in range(len(self.system_dict["dataset"]["params"]["classes"])):
            class_dict[self.system_dict["dataset"]["params"]["classes"][i]] = {};
            class_dict[self.system_dict["dataset"]["params"]["classes"][i]]["num_labels"] = 0;
            class_dict[self.system_dict["dataset"]["params"]["classes"][i]]["num_correct"] = 0;


        for inputs, labels in self.system_dict["local"]["data_loaders"]["test"]:
            if(self.system_dict["training"]["settings"]["display_progress_realtime"] and self.system_dict["verbose"]):
                pbar.update();

            inputs = inputs.to(self.system_dict["local"]["device"]);
            labels = labels.to(self.system_dict["local"]["device"]);
            labels = labels.cpu().detach().numpy()[0]

            outputs = self.system_dict["local"]["model"](inputs);

            list_classes = [];
            list_labels = [];
            raw_scores = outputs.cpu().detach().numpy()[0];

            for i in range(len(raw_scores)):
                prob = logistic.cdf(raw_scores[i])
                if(prob > 0.5):
                    list_classes.append(self.system_dict["dataset"]["params"]["classes"][i])

            for i in range(len(labels)):
                if(labels[i]):
                    list_labels.append(self.system_dict["dataset"]["params"]["classes"][i])

            for i in range(len(list_labels)):
                actual = list_labels[i];
                if actual in list_classes:
                    correct = True;
                else:
                    correct = False;

                index = self.system_dict["dataset"]["params"]["classes"].index(actual);

                class_dict[self.system_dict["dataset"]["params"]["classes"][index]]["num_labels"] += 1;
                total_labels += 1;

                if(correct):
                    class_dict[self.system_dict["dataset"]["params"]["classes"][index]]["num_correct"] += 1;
                    running_corrects += 1;

            

        accuracy = running_corrects/total_labels;

        self.custom_print("");
        self.custom_print("    Result");
        self.custom_print("        class based accuracies");
        for i in range(len(self.system_dict["dataset"]["params"]["classes"])):
            self.custom_print("            {}. {} - {} %".format(i, self.system_dict["dataset"]["params"]["classes"][i], 
                class_dict[self.system_dict["dataset"]["params"]["classes"][i]]["num_correct"]/class_dict[self.system_dict["dataset"]["params"]["classes"][i]]["num_labels"]*100));
            class_dict[self.system_dict["dataset"]["params"]["classes"][i]]["accuracy(%)"] = class_dict[self.system_dict["dataset"]["params"]["classes"][i]]["num_correct"]/class_dict[self.system_dict["dataset"]["params"]["classes"][i]]["num_labels"]*100;
        self.custom_print("        total labels:            {}".format(total_labels));
        self.custom_print("        num correct predictions: {}".format(running_corrects));
        self.custom_print("        Average accuracy (%):    {}".format(accuracy*100));
        self.system_dict["testing"]["class_accuracy"] = class_dict
        self.system_dict["testing"]["status"] = True;
        self.custom_print("");
        return accuracy*100, class_dict;

    ###############################################################################################################################################





    ###############################################################################################################################################
    @accepts("self", img_name=[str, bool], img_dir=[str, bool], return_raw=bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_prediction_final(self, img_name=False, img_dir=False, return_raw=False):
        '''
        Main function for external inferencing on single image or folder of images post training

        Args:
            img_name (str): path to image
            img_dir (str): path to folders containing images. 
                            (Optional)
            return_raw (bool): If True, then output dictionary contains image probability for every class in the set.
                                Else, only the most probable class score is returned back.
                                

        Returns:
            float: Accuracy in percentage
            dict: Inference output
                   1) Image name
                   2) Predicted class
                   3) Predicted score
        '''
        self.custom_print("Prediction");

        if(not self.system_dict["dataset"]["params"]["input_size"]):
            msg = "Input Size not set for experiment.\n";
            msg += "Tip: Use update_input_size";
            raise ConstraintError(msg);

        self.system_dict = set_transform_test(self.system_dict);


        if(not self.system_dict["dataset"]["params"]["classes"]):
            msg = "Class information unavailabe.\n";
            msg += "Labels returned - Indexes instead of classes";
            ConstraintWarning(msg);


        if(img_name):
            self.custom_print("    Image name:         {}".format(img_name));

            label, score, raw_output = process_single(img_name, return_raw, self.system_dict);

            self.custom_print("    Predicted class:      {}".format(label));
            self.custom_print("    Predicted score:      {}".format(score));
            tmp = {};
            tmp["img_name"] = img_name;
            tmp["predicted_class"] = label;
            tmp["score"] = score;
            if(return_raw):
                tmp["raw"] = raw_output;
            self.custom_print("");                
            return tmp;

        if(img_dir):
            output = [];
            self.custom_print("    Dir path:           {}".format(img_dir));
            img_list = os.listdir(img_dir);
            self.custom_print("    Total Images:       {}".format(len(img_list)));
            self.custom_print("Processing Images");
            if(self.system_dict["verbose"]):
                pbar = tqdm(total=len(img_list));


            for i in range(len(img_list)):
                if(self.system_dict["verbose"]):
                    pbar.update();

                img_name = img_dir + "/" + img_list[i];
                
                label, score, raw_output = process_single(img_name, return_raw, self.system_dict);

                tmp = {};
                tmp["img_name"] = img_list[i];
                tmp["predicted_class"] = label;
                tmp["score"] = score;
                if(return_raw):
                    tmp["raw"] = raw_output;
                output.append(tmp);
            self.custom_print("");

            return output
    ###############################################################################################################################################




    ###############################################################################################################################################
    @accepts("self", img_name=[str, bool], img_dir=[str, bool], return_raw=bool, img_thresh=float, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_prediction_final_multiple(self, img_name=False, img_dir=False, return_raw=False, img_thresh=0.5):
        '''
        Main function for external inferencing on single image or folder of images post training
        - For multi-label image classification

        Args:
            img_name (str): path to image
            img_dir (str): path to folders containing images. 
                            (Optional)
            return_raw (bool): If True, then output dictionary contains image probability for every class in the set.
                                Else, only the most probable class score is returned back.
            img_thresh (float): Thresholding for multi label image classification.
                                

        Returns:
            float: Accuracy in percentage
            dict: Inference output
                   1) Image name
                   2) Predicted classes list
                   3) Predicted score
        '''
        self.custom_print("Prediction");

        if(not self.system_dict["dataset"]["params"]["input_size"]):
            msg = "Input Size not set for experiment.\n";
            msg += "Tip: Use update_input_size";
            raise ConstraintError(msg);

        self.system_dict = set_transform_test(self.system_dict);


        if(not self.system_dict["dataset"]["params"]["classes"]):
            msg = "Class information unavailabe.\n";
            msg += "Labels returned - Indexes instead of classes";
            ConstraintWarning(msg);


        if(img_name):
            self.custom_print("    Image name:         {}".format(img_name));

            labels, scores, raw_output = process_multi(img_name, return_raw, img_thresh, self.system_dict);

            self.custom_print("    Predicted classes:      {}".format(labels));
            self.custom_print("    Predicted scorees:      {}".format(scores));
            tmp = {};
            tmp["img_name"] = img_name;
            tmp["predicted_classes"] = labels;
            tmp["scores"] = scores;
            if(return_raw):
                tmp["raw"] = raw_output;
            self.custom_print("");                
            return tmp;

        
        if(img_dir):
            output = [];
            self.custom_print("    Dir path:           {}".format(img_dir));
            img_list = os.listdir(img_dir);
            self.custom_print("    Total Images:       {}".format(len(img_list)));
            self.custom_print("Processing Images");
            if(self.system_dict["verbose"]):
                pbar = tqdm(total=len(img_list));


            for i in range(len(img_list)):
                if(self.system_dict["verbose"]):
                    pbar.update();

                img_name = img_dir + "/" + img_list[i];
                
                labels, scores, raw_output = process_multi(img_name, return_raw, img_thresh, self.system_dict);

                tmp = {};
                tmp["img_name"] = img_list[i];
                tmp["predicted_classes"] = labels;
                tmp["scores"] = scores;
                if(return_raw):
                    tmp["raw"] = raw_output;
                output.append(tmp);
            self.custom_print("");

            return output
        
    ###############################################################################################################################################


from monk.gluon.finetune.imports import *
from monk.system.imports import *

from monk.gluon.finetune.level_13_updates_main import prototype_updates



class prototype_master(prototype_updates):
    '''
    Main class for all functions in expert mode

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''
    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose);
	
	###############################################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def Dataset(self):
        '''
        Load transforms and set dataloader

        Args:
            None

        Returns:
            None
        '''
        self.set_dataset_final(test=self.system_dict["states"]["eval_infer"]);
        save(self.system_dict);

        
        if(self.system_dict["states"]["eval_infer"]):
            
            self.custom_print("Pre-Composed Test Transforms");
            self.custom_print(self.system_dict["dataset"]["transforms"]["test"]);
            self.custom_print("");

            self.custom_print("Dataset Numbers");
            self.custom_print("    Num test images: {}".format(self.system_dict["dataset"]["params"]["num_test_images"]));
            self.custom_print("    Num classes:      {}".format(self.system_dict["dataset"]["params"]["num_classes"]))
            self.custom_print("");

        else:
            
            self.custom_print("Pre-Composed Train Transforms");
            self.custom_print(self.system_dict["dataset"]["transforms"]["train"]);
            self.custom_print("");
            self.custom_print("Pre-Composed Val Transforms");
            self.custom_print(self.system_dict["dataset"]["transforms"]["val"]);
            self.custom_print("");

            self.custom_print("Dataset Numbers");
            self.custom_print("    Num train images: {}".format(self.system_dict["dataset"]["params"]["num_train_images"]));
            self.custom_print("    Num val images:   {}".format(self.system_dict["dataset"]["params"]["num_val_images"]));
            self.custom_print("    Num classes:      {}".format(self.system_dict["dataset"]["params"]["num_classes"]))
            self.custom_print("");
        
    ###############################################################################################################################################


    ###############################################################################################################################################
    @accepts("self", [int, float], post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def Dataset_Percent(self, percent):
        '''
        Select a portion of dataset

        Args:
            percent (bool): percentage of sub-dataset

        Returns:
            None
        '''
        sampled_dataset = None;
        image_datasets = {};
        dataset_type = self.system_dict["dataset"]["dataset_type"];
        dataset_train_path = self.system_dict["dataset"]["train_path"];
        dataset_val_path = self.system_dict["dataset"]["val_path"];
        csv_train = self.system_dict["dataset"]["csv_train"];
        csv_val = self.system_dict["dataset"]["csv_val"];
        train_val_split = self.system_dict["dataset"]["params"]["train_val_split"];
        delimiter = self.system_dict["dataset"]["params"]["delimiter"];
        batch_size = self.system_dict["dataset"]["params"]["batch_size"];
        shuffle = self.system_dict["dataset"]["params"]["train_shuffle"];
        num_workers = self.system_dict["dataset"]["params"]["num_workers"];

        
        if(dataset_type == "train"):
            label_list = [];
            image_list = [];
            classes = os.listdir(dataset_train_path);
            for i in range(len(classes)):
                tmp_image_list = os.listdir(dataset_train_path + "/" + classes[i]);
                subset_image_list = tmp_image_list[:int(len(tmp_image_list)*percent/100.0)];
                result = list(map(lambda x: classes[i] + "/" + x, subset_image_list))
                tmp_label_list = [classes[i]]*len(subset_image_list);
                label_list += tmp_label_list;
                image_list += result;
            image_label_dict = {'ID': image_list, 'Label': label_list}  
            df = pd.DataFrame(image_label_dict);
            df.to_csv("sampled_dataset_train.csv", index=False);
        elif(dataset_type == "train-val"):
            label_list = [];
            image_list = [];
            classes = os.listdir(dataset_train_path);
            for i in range(len(classes)):
                tmp_image_list = os.listdir(dataset_train_path + "/" + classes[i]);
                subset_image_list = tmp_image_list[:int(len(tmp_image_list)*percent/100.0)];
                result = list(map(lambda x: classes[i] + "/" + x, subset_image_list))
                tmp_label_list = [classes[i]]*len(subset_image_list);
                label_list += tmp_label_list;
                image_list += result;
            image_label_dict = {'ID': image_list, 'Label': label_list}  
            df = pd.DataFrame(image_label_dict);
            df.to_csv("sampled_dataset_train.csv", index=False);

            label_list = [];
            image_list = [];
            classes = os.listdir(dataset_train_path);
            for i in range(len(classes)):
                tmp_image_list = os.listdir(dataset_val_path + "/" + classes[i]);
                subset_image_list = tmp_image_list[:int(len(tmp_image_list)*percent/100.0)];
                result = list(map(lambda x: classes[i] + "/" + x, subset_image_list))
                tmp_label_list = [classes[i]]*len(subset_image_list);
                label_list += tmp_label_list;
                image_list += result;
            image_label_dict = {'ID': image_list, 'Label': label_list}  
            df = pd.DataFrame(image_label_dict);
            df.to_csv("sampled_dataset_val.csv", index=False);
        elif(dataset_type == "csv_train"):
            df = pd.read_csv(csv_train);
            df = df.iloc[np.random.permutation(len(df))]
            df_sampled = df.iloc[:int(len(df)*percent/100.0)];
            df_sampled.to_csv("sampled_dataset_train.csv", index=False);
        elif(dataset_type == "csv_train-val"):
            df = pd.read_csv(csv_train);
            df = df.iloc[np.random.permutation(len(df))]
            df_sampled = df.iloc[:int(len(df)*percent/100.0)];
            df_sampled.to_csv("sampled_dataset_train.csv", index=False);
            df = pd.read_csv(csv_val);
            df = df.iloc[np.random.permutation(len(df))]
            df_sampled = df.iloc[:int(len(df)*percent/100.0)];
            df_sampled.to_csv("sampled_dataset_val.csv", index=False);



    ###############################################################################################################################################





    ###############################################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def Model(self):
        '''
        Load Model as per paraameters set

        Args:
            None

        Returns:
            None
        '''
        if(self.system_dict["states"]["copy_from"]):
            msg = "Cannot set model in Copy-From mode.\n";
            raise ConstraintError(msg)
        self.set_model_final();
        save(self.system_dict)
    ###############################################################################################################################################




    ###############################################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def Train(self):
        '''
        Master function for training

        Args:
            None

        Returns:
            None
        '''
        self.set_training_final();
        save(self.system_dict);
    ###############################################################################################################################################



    ###############################################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def Evaluate(self):
        '''
        Master function for external validation

        Args:
            None

        Returns:
            None
        '''
        if(self.system_dict["dataset"]["label_type"] == "single" or self.system_dict["dataset"]["label_type"] == False):
            accuracy, class_based_accuracy = self.set_evaluation_final();
            save(self.system_dict);
        else:
            accuracy, class_based_accuracy = self.set_evaluation_final_multiple();
            save(self.system_dict);
        return accuracy, class_based_accuracy;
    ###############################################################################################################################################


    ###############################################################################################################################################
    @error_checks(None, img_name=["file", "r"], img_dir=["folder", "r"], return_raw=False, img_thresh=["gte", 0.0, "lte", 1.0], post_trace=False)
    @accepts("self", img_name=[str, bool], img_dir=[str, bool], return_raw=bool, img_thresh=float, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def Infer(self, img_name=False, img_dir=False, return_raw=False, img_thresh=0.5):
        '''
        Master function for inference 

        Args:
            img_name (str): path to image
            img_dir (str): path to folders containing images. 
                            (Optional)
            return_raw (bool): If True, then output dictionary contains image probability for every class in the set.
                                Else, only the most probable class score is returned back.
            img_thresh (float): Thresholding for multi label image classification.

        Returns:
            dict: Dictionary containing details on predictions.
        '''
        if(self.system_dict["dataset"]["label_type"] == "single" or self.system_dict["dataset"]["label_type"] == False):
            if(not img_dir):
                predictions = self.set_prediction_final(img_name=img_name, return_raw=return_raw);
            else:
                predictions = self.set_prediction_final(img_dir=img_dir, return_raw=return_raw);
            return predictions;
        else:
            if(not img_dir):
                predictions = self.set_prediction_final_multiple(img_name=img_name, return_raw=return_raw, img_thresh=img_thresh);
            else:
                predictions = self.set_prediction_final_multiple(img_dir=img_dir, return_raw=return_raw, img_thresh=img_thresh);
            return predictions;
    ###############################################################################################################################################



    ###############################################################################################################################################
    @accepts("self", network=list, data_shape=tuple, use_gpu=bool, network_initializer=str, debug=bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def Compile_Network(self, network, data_shape=(3, 224, 224), use_gpu=True, network_initializer="xavier_normal", debug=True):
        '''
        Master function for compiling custom network and initializing it 

        Args:
            network: Network stacked as list of lists
            data_shape (tuple): Input shape of data in format C, H, W
            use_gpu (bool): If True, model loaded on gpu
            network_initializer (str): Initialize network with random weights. Select the random generator type function.

        Returns:
            None
        '''
        self.system_dict["custom_model"]["network_stack"] = network;
        self.system_dict["custom_model"]["network_initializer"] = network_initializer;
        self.system_dict["model"]["type"] = "custom";
        self.system_dict["dataset"]["params"]["data_shape"] = data_shape;
        self.system_dict["custom_model"]["debug"] = debug;
        self.system_dict = set_device(use_gpu, self.system_dict);
        save(self.system_dict);
        self.set_model_final();
    ###############################################################################################################################################


    ###############################################################################################################################################
    @accepts("self", data_shape=tuple, port=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def Visualize_With_Netron(self, data_shape=None, port=None):
        '''
        Visualize network with netron library 

        Args:
            data_shape (tuple): Input shape of data in format C, H, W
            port (int): Local host free port.

        Returns:
            None
        '''
        self.custom_print("Using Netron To Visualize");
        self.custom_print("Not compatible on kaggle");
        self.custom_print("Compatible only for Jupyter Notebooks");

        if not data_shape:
            c, h, w = self.system_dict["dataset"]["params"]["data_shape"];
        else:
            c, h, w = data_shape;

        data = mx.nd.random.randn(1, c, h, w)
        if(self.system_dict["model"]["params"]["use_gpu"]):
            self.system_dict["local"]["ctx"] = [mx.gpu(0)];
        else:
            self.system_dict["local"]["ctx"] = [mx.cpu()];

        data = data.copyto(self.system_dict["local"]["ctx"][0])

        self.system_dict["local"]["model"].hybridize();
        out = self.system_dict["local"]["model"](data);

        self.system_dict["local"]["model"].export("model", epoch=0)

        import netron
        if(not port):
            netron.start('model-symbol.json')
        else:
            netron.start('model-symbol.json', port=port)
    ###############################################################################################################################################


from monk.system.imports import *
from monk.system.base_system_state import get_base_system_dict
from monk.system.common import read_json
from monk.system.common import save
from monk.system.common import create_dir
from monk.system.common import delete_dir



class system():
    '''
    Base class for all system project management

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''
    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        self.system_dict = get_base_system_dict();
        self.system_dict["verbose"] = verbose;
        self.system_dict["cwd"] = os.getcwd() + "/";
        self.system_dict["master_systems_dir"] = self.system_dict["cwd"] + "workspace/";
        self.system_dict["master_systems_dir_relative"] = "workspace/";

        create_dir(self.system_dict["master_systems_dir_relative"]);

        self.system_dict["master_comparison_dir"] = self.system_dict["cwd"] + "workspace/comparison/";
        self.system_dict["master_comparison_dir_relative"] = "workspace/comparison/";
        
        create_dir(self.system_dict["master_comparison_dir_relative"]);
            
        self.system_dict["local"]["projects_list"] = os.listdir(self.system_dict["master_comparison_dir"]);
        self.system_dict["local"]["num_projects"] = len(self.system_dict["local"]["projects_list"]);
        self.system_dict["local"]["experiments_list"] = [];
        self.system_dict["local"]["num_experiments"] = 0;
        self.system_dict["origin"] = ["New", "New"];



    #############################################################################################################################
    @accepts("self", str, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_system_project(self, project_name):
        '''
        Create Project

        Args:
            project_name (str): Unique name to project

        Returns:
            None
        '''
        self.system_dict["project_dir"] = self.system_dict["master_systems_dir"] + project_name + "/";
        self.system_dict["project_dir_relative"] = self.system_dict["master_systems_dir_relative"] + project_name + "/";
        if(not os.path.isdir(self.system_dict["project_dir"])):
            self.system_dict["local"]["projects_list"].append(project_name);
            self.system_dict["local"]["num_projects"] += 1;
        create_dir(self.system_dict["project_dir"]);
        self.set_system_select_project(project_name);
    
    
    @accepts("self", str, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_system_select_project(self, project_name):
        '''
        Function to update system dictionary on project properties

        Args:
            project_name (str): Unique name to project

        Returns:
            None
        '''
        self.system_dict["project_name"] = project_name;
        self.system_dict["local"]["experiments_list"] = os.listdir(self.system_dict["project_dir"]);
        self.system_dict["local"]["num_experiments"] = len(self.system_dict["local"]["experiments_list"]);


    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_system_aux_list_projects(self):
        '''
        List all projects in current workspace

        Args:
            None

        Returns:
            None
        '''
        return os.listdir(self.system_dict["master_systems_dir"]);
    #############################################################################################################################



    #############################################################################################################################
    @accepts("self", str, eval_infer=bool, copy_from=[list, bool], pseudo_copy_from=[list, bool], resume_train=bool, summary=bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_system_experiment(self, experiment_name, eval_infer=False, copy_from=False, pseudo_copy_from=False, resume_train=False, summary=False):
        '''
        Create Experiment or load it in different states

        Args:
            experiment_name (str): Unique name to experiment
            eval_infer (bool): If set as True, model is loaded in evaluation mode
            resume_train (bool): If set as True, model is loaded from last checkpoint
            copy_from (list): [project, experiment] to copy from
            pseudo_copy_from (list): For creating sub-experiments while in hyper-parametric analysis state
            summary (list): Dummy variable

        Returns:
            None
        '''
        if(summary):
            self.set_system_select_experiment(experiment_name);
            print_summary(self.system_dict["fname_relative"]);

        else:
            self.system_dict["experiment_dir"] = self.system_dict["project_dir"] + experiment_name + "/";
            self.system_dict["experiment_dir_relative"] = self.system_dict["project_dir_relative"] + experiment_name + "/";
            if(not os.path.isdir(self.system_dict["experiment_dir"])):
                self.system_dict["local"]["experiments_list"].append(experiment_name);
                self.system_dict["local"]["num_experiments"] += 1;
            create_dir(self.system_dict["experiment_dir"]);
            self.set_system_select_experiment(experiment_name);
            
            if(eval_infer):
                self.set_system_state_eval_infer();
            elif(resume_train):
                self.set_system_state_resume_train();
            elif(copy_from):
                self.set_system_delete_create_dir();
                self.set_system_state_copy_from(copy_from);
            elif(pseudo_copy_from):
                self.set_system_delete_create_dir();
                self.set_system_state_pseudo_copy_from(pseudo_copy_from);
            else: 
                self.set_system_delete_create_dir();
                save(self.system_dict);

    
    @accepts("self", str, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_system_select_experiment(self, experiment_name):
        '''
        Function to update system dictionary on experiment properties

        Args:
            experiment_name (str): Unique name to experiment

        Returns:
            None
        '''
        self.system_dict["experiment_name"] = experiment_name;
        self.system_dict["output_dir"] = self.system_dict["experiment_dir"] + "output/";
        self.system_dict["output_dir_relative"] = self.system_dict["experiment_dir_relative"] + "output/";
        self.system_dict["model_dir"] = self.system_dict["output_dir"] + "models/";
        self.system_dict["model_dir_relative"] = self.system_dict["output_dir_relative"] + "models/";
        self.system_dict["log_dir"] = self.system_dict["output_dir"] + "logs/";
        self.system_dict["log_dir_relative"] = self.system_dict["output_dir_relative"] + "logs/";
        self.system_dict["fname"] = self.system_dict["experiment_dir"] + "/experiment_state.json";
        self.system_dict["fname_relative"] = self.system_dict["experiment_dir_relative"] + "/experiment_state.json";
        self.system_dict["visualization"]["base"] = self.system_dict["experiment_dir"] + "visualization/";
        self.system_dict["visualization"]["kernels_dir"] = self.system_dict["visualization"]["base"] + "kernels/";
        self.system_dict["visualization"]["feature_maps_dir"] = self.system_dict["visualization"]["base"] + "feature_maps/";

    #############################################################################################################################



    #############################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_system_delete_create_dir(self):
        '''
        Function to remove old directories and create new at the same place

        Args:
            None

        Returns:
            None
        '''
        delete_dir(self.system_dict["output_dir_relative"]);
        create_dir(self.system_dict["output_dir_relative"]);
        create_dir(self.system_dict["model_dir_relative"]);
        create_dir(self.system_dict["log_dir_relative"]);
    #############################################################################################################################




    #############################################################################################################################
    @accepts("self", str, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_system_comparison(self, comparison_name):
        '''
        Create comparison experiment

        Args:
            comparison_name (str): Unique name to comparison experiment

        Returns:
            None
        '''
        create_dir(self.system_dict["master_comparison_dir"] + comparison_name + "/");
        self.system_dict["comparison_name"] = comparison_name;
        self.system_dict["comparison_dir"] = self.system_dict["master_comparison_dir"] + comparison_name + "/";
    #############################################################################################################################


    #############################################################################################################################
    @accepts("self", [str, int, list, dict, float, tuple], post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def custom_print(self, msg):
        '''
        Overwritten print function, to switch off and on as per verbosity levels

        Args:
            msg (str): Message to print

        Returns:
            None
        '''
        if(self.system_dict["verbose"]):
            print(msg);
    #############################################################################################################################






    #############################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def print_list_models(self):
        '''
        List all the available models in the selected backend.

        Args:
            None

        Returns:
            None
        '''
        self.custom_print("Models List: ");

        if(self.system_dict["library"] == "Mxnet"):
            set1 = ["alexnet", "darknet53", "DenseNet121", "DenseNet161", "DenseNet169", "DenseNet201", "InceptionV3", "MobileNet1.0", "MobileNet0.75", 
                        "MobileNet0.25", "MobileNet0.5", "ResNet18_v1", "ResNet34_v1", "ResNet50_v1", "ResNet101_v1", "ResNet152_v1", "ResNext50_32x4d", 
                        "ResNext101_32x4d", "ResNext101_64x4d_v1", "SE_ResNext50_32x4d", "SE_ResNext101_32x4d", "SE_ResNext101_64x4d", "SENet_154", 
                        "VGG11", "VGG13", "VGG16", "VGG19", "VGG11_bn", "VGG13_bn", "VGG16_bn", "VGG19_bn", "ResNet18_v2", "ResNet34_v2", 
                        "ResNet50_v2", "ResNet101_v2", "ResNet152_v2"];
            set2 = ["MobileNetV2_1.0", "MobileNetV2_0.75", "MobileNetV2_0.5", "MobileNetV2_0.25", "SqueezeNet1.0", "SqueezeNet1.1", "MobileNetV3_Large", "MobileNetV3_Small"];
            set3 = ["ResNet18_v1b", "ResNet34_v1b", "ResNet50_v1b", "ResNet50_v1b_gn", "ResNet101_v1b", "ResNet152_v1b", "ResNet50_v1c", 
                        "ResNet101_v1c", "ResNet152_v1c", "ResNet50_v1d", "ResNet101_v1d", "ResNet152_v1d", "ResNet18_v1d", "ResNet34_v1d", 
                        "ResNet50_v1d", "ResNet101_v1d", "ResNet152_v1d", "resnet18_v1b_0.89", "resnet50_v1d_0.86", "resnet50_v1d_0.48", 
                        "resnet50_v1d_0.37", "resnet50_v1d_0.11", "resnet101_v1d_0.76", "resnet101_v1d_0.73", "Xception"];
            combined_list = set1+set2+set3
            combined_list_lower = list(map(str.lower, combined_list))


        elif(self.system_dict["library"] == "Pytorch"):
            set1 = ["alexnet", "vgg11", "vgg11_bn", "vgg13", "vgg13_bn", "vgg16", "vgg16_bn", "vgg19", "vgg19_bn"]
            set2 = ["densenet121", "densenet161", "densenet169", "densenet201"]
            set3 = ["googlenet", "inception_v3", "resnet18", "resnet34", "resnet50", "resnet101", "resnet152", "resnext50_32x4d", "resnext101_32x8d",
                        "shufflenet_v2_x0_5", "shufflenet_v2_x1_0, shufflenet_v2_x1_5", "shufflenet_v2_x2_0", "wide_resnet101_2", "wide_resnet50_2"]
            set4 = ["mnasnet0_5", "mnasnet0_75", "mnasnet1_0", "mnasnet1_3", "mobilenet_v2", "squeezenet1_0", "squeezenet1_1"]
            combined_list = set1+set2+set3+set4
            combined_list_lower = list(map(str.lower, combined_list))


        elif(self.system_dict["library"] == "Keras"):
            set1 = ["mobilenet", "densenet121", "densenet169", "densenet201", "inception_v3", 
                        "inception_resnet_v3", "mobilenet_v2", "nasnet_mobile", "nasnet_large", "resnet50",
                        "resnet101", "resnet152", "resnet50_v2", "resnet101_v2", "resnet152_v2", "vgg16",
                        "vgg19", "xception"];
            combined_list = set1
            combined_list_lower = list(map(str.lower, combined_list))
            
        for i in range(len(combined_list_lower)):
            self.custom_print("    {}. {}".format(i+1, combined_list_lower[i]))
        self.custom_print("")
    #############################################################################################################################




    #############################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def print_list_layers_transfer_learning(self):
        '''
        List all the available transfer learning layers in the selected backend.

        Args:
            None

        Returns:
            None
        '''
        self.custom_print("Layers List for transfer learning: ");

        if(self.system_dict["library"] == "Mxnet"):
            combined_list_lower = ["append_linear", "append_dropout"];

        elif(self.system_dict["library"] == "Keras"):
            combined_list_lower = ["append_linear", "append_dropout"];

        elif(self.system_dict["library"] == "Pytorch"):
            combined_list_lower = ["append_linear", "append_dropout"];

        for i in range(len(combined_list_lower)):
            self.custom_print("    {}. {}".format(i+1, combined_list_lower[i]))

        self.custom_print("")
    #############################################################################################################################




    #############################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def print_list_layers_custom_model(self):
        '''
        List all the available custom network layers in the selected backend.

        Args:
            None

        Returns:
            None
        '''
        self.custom_print("Layers List for transfer learning: ");

        if(self.system_dict["library"] == "Mxnet"):
            combined_list_lower = ["convolution1d", "convolution2d", "convolution", "convolution3d", "transposed_convolution1d",
                                    "transposed_convolution", "transposed_convolution2d", "transposed_convolution3d", 
                                    "max_pooling1d", "max_pooling2d", "max_pooling", "max_pooling3d", "average_pooling1d",
                                    "average_pooling2d", "average_pooling", "average_pooling3d", "global_max_pooling1d",
                                    "global_max_pooling2d", "global_max_pooling", "global_max_pooling3d", "global_average_pooling1d",
                                    "global_average_pooling2d", "global_average_pooling", "global_average_pooling3d", 
                                    "fully_connected", "dropout", "flatten", "identity", "add", "concatenate", "batch_normalization",
                                    "instance_normalization", "layer_normalization"];

        elif(self.system_dict["library"] == "Keras"):
            combined_list_lower = ["convolution1d", "convolution2d", "convolution", "convolution3d", "transposed_convolution", 
                                    "transposed_convolution2d", "transposed_convolution3d", "max_pooling1d", "max_pooling2d", 
                                    "max_pooling", "max_pooling3d", "average_pooling1d", "average_pooling2d", "average_pooling", 
                                    "average_pooling3d", "global_max_pooling1d", "global_max_pooling2d", "global_max_pooling", 
                                    "global_max_pooling3d", "global_average_pooling1d", "global_average_pooling2d", "global_average_pooling", 
                                    "global_average_pooling3d", "flatten", "fully_connected", "dropout", "identity", "batch_normalization",
                                    "add", "concatenate"];

        elif(self.system_dict["library"] == "Pytorch"):
            combined_list_lower = ["convolution1d", "convolution2d", "convolution", "convolution3d", "transposed_convolution1d",
                                    "transposed_convolution", "transposed_convolution2d", "transposed_convolution3d",
                                    "max_pooling1d", "max_pooling2d", "max_pooling", "max_pooling3d", "average_pooling1d",
                                    "average_pooling2d", "average_pooling", "average_pooling3d", "global_max_pooling1d",
                                    "global_max_pooling2d", "global_max_pooling", "global_max_pooling3d", "global_average_pooling1d",
                                    "global_average_pooling2d", "global_average_pooling", "global_average_pooling3d", "fully_connected", 
                                    "flatten", "dropout", "identity", "batch_normalization", "instance_normalization", "layer_normalization",
                                    "add", "concatenate"];

        for i in range(len(combined_list_lower)):
            self.custom_print("    {}. {}".format(i+1, combined_list_lower[i]))

        self.custom_print("")
    #############################################################################################################################





    #############################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def print_list_activations_transfer_learning(self):
        '''
        List all the available transfer learning activations in the selected backend.

        Args:
            None

        Returns:
            None
        '''
        self.custom_print("Activations List for transfer learning: ");

        if(self.system_dict["library"] == "Mxnet"):
            combined_list_lower = ["append_elu", "append_leakyrelu", "append_prelu", "append_relu", "append_selu",
                                    "append_selu", "append_sigmoid", "append_softplus", "append_tanh",
                                    "append_softmax", "append_swish"];

        elif(self.system_dict["library"] == "Keras"):
            combined_list_lower = ["append_elu", "append_leakyrelu", "append_prelu", "append_relu", "append_selu",
                                    "append_selu", "append_sigmoid", "append_softplus", "append_softsign", "append_tanh",
                                    "append_threshold", "append_softmax"];



        elif(self.system_dict["library"] == "Pytorch"):
            combined_list_lower = ["append_elu", "append_leakyrelu", "append_prelu", "append_relu", "append_selu",
                                    "append_selu", "append_sigmoid", "append_softplus", "append_softsign", "append_tanh",
                                    "append_threshold", "append_softmax", "append_hardshrink", "append_hardtanh", 
                                    "append_logsigmoid", "append_relu6", "append_rrelu", "append_celu", "append_softshrink",
                                    "append_tanhshrink", "append_logsoftmax", "append_softmin"];


        for i in range(len(combined_list_lower)):
            self.custom_print("    {}. {}".format(i+1, combined_list_lower[i]))

        self.custom_print("")
    #############################################################################################################################




    #############################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def print_list_activations_custom_model(self):
        '''
        List all the available custom network activations in the selected backend.

        Args:
            None

        Returns:
            None
        '''
        self.custom_print("Activations List for transfer learning: ");

        if(self.system_dict["library"] == "Mxnet"):
            combined_list_lower = ["relu", "sigmoid", "tanh", "softplus", "softsign", "elu", "gelu", "leaky_relu",
                                    "prelu", "selu", "swish"];

        elif(self.system_dict["library"] == "Keras"):
            combined_list_lower = ["relu", "elu", "leaky_relu", "prelu", "thresholded_relu", "softmax", 
                                    "selu", "softplus", "softsign", "tanh", "sigmoid", "hard_sigmoid"];



        elif(self.system_dict["library"] == "Pytorch"):
            combined_list_lower = ["relu", "sigmoid", "tanh", "softplus", "softsign",  "elu", "leaky_relu", "prelu", "selu",
                                    "hardshrink", "hardtanh", "logsigmoid", "relu6", "rrelu", "celu", "softshrink", "tanhshrink",
                                    "threshold", "softmin", "softmax", "logsoftmax"];


        for i in range(len(combined_list_lower)):
            self.custom_print("    {}. {}".format(i+1, combined_list_lower[i]))

        self.custom_print("")
    #############################################################################################################################






    #############################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def print_list_losses(self):
        '''
        List all the available loss functions in the selected backend.

        Args:
            None

        Returns:
            None
        '''
        self.custom_print("Losses List: ");

        if(self.system_dict["library"] == "Mxnet"):
            combined_list_lower = ["loss_l1", "loss_l2", "loss_softmax_crossentropy", "loss_crossentropy",
                                    "loss_sigmoid_binary_crossentropy", "loss_binary_crossentropy",
                                    "loss_kldiv", "loss_poisson_nll", "loss_huber", "loss_hinge",
                                    "loss_squared_hinge"];

        elif(self.system_dict["library"] == "Keras"):
            combined_list_lower = ["loss_l1", "loss_l2", "loss_crossentropy", "loss_binary_crossentropy", 
                                    "loss_kldiv", "loss_hinge", "loss_squared_hinge"];


        elif(self.system_dict["library"] == "Pytorch"):
            combined_list_lower = ["loss_l1", "loss_l2", "loss_softmax_crossentropy", "loss_crossentropy",
                                    "loss_sigmoid_binary_crossentropy", "loss_binary_crossentropy",
                                    "loss_kldiv", "loss_poisson_nll", "loss_huber", "loss_hinge",
                                    "loss_squared_hinge", "loss_multimargin", "loss_squared_multimargin",
                                    "loss_multilabel_margin", "loss_multilabel_softmargin"];

                                    

        for i in range(len(combined_list_lower)):
            self.custom_print("    {}. {}".format(i+1, combined_list_lower[i]))

        self.custom_print("")
    #############################################################################################################################





    #############################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def print_list_optimizers(self):
        '''
        List all the available optimizers in the selected backend.

        Args:
            None

        Returns:
            None
        '''
        self.custom_print("Optimizers List: ");

        if(self.system_dict["library"] == "Mxnet"):
            combined_list_lower = ["optimizer_sgd", "optimizer_nesterov_sgd", "optimizer_rmsprop", "optimizer_momentum_rmsprop", 
                                    "optimizer_adam", "optimizer_adagrad", "optimizer_nesterov_adam", 
                                    "optimizer_adadelta", "optimizer_adamax", "optimizer_signum"];

        elif(self.system_dict["library"] == "Keras"):
            combined_list_lower = ["optimizer_sgd", "optimizer_nesterov_sgd", "optimizer_rmsprop", "optimizer_adam",
                                    "optimizer_nesterov_adam", "optimizer_adamax", "optimizer_adagrad", "optimizer_adadelta"];


        elif(self.system_dict["library"] == "Pytorch"):
            combined_list_lower = ["optimizer_sgd", "optimizer_nesterov_sgd", "optimizer_rmsprop", "optimizer_momentum_rmsprop", 
                                    "optimizer_adam", "optimizer_adamax", "optimizer_adamw", "optimizer_adagrad", 
                                    "optimizer_adadelta"];

                                    
        for i in range(len(combined_list_lower)):
            self.custom_print("    {}. {}".format(i+1, combined_list_lower[i]))

        self.custom_print("")
    #############################################################################################################################





    #############################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def print_list_schedulers(self):
        '''
        List all the available learning rate schedulers in the selected backend.

        Args:
            None

        Returns:
            None
        '''
        self.custom_print("Optimizers List: ");

        if(self.system_dict["library"] == "Mxnet"):
            combined_list_lower = ["lr_fixed", "lr_step_decrease", "lr_multistep_decrease"];

        elif(self.system_dict["library"] == "Keras"):
            combined_list_lower = ["lr_fixed", "lr_step_decrease", "lr_exponential_decrease", "lr_plateau_decrease"];


        elif(self.system_dict["library"] == "Pytorch"):
            combined_list_lower = ["lr_fixed", "lr_step_decrease", "lr_multistep_decrease", "lr_exponential_decrease", "lr_plateau_decrease"];

                                    
        for i in range(len(combined_list_lower)):
            self.custom_print("    {}. {}".format(i+1, combined_list_lower[i]))

        self.custom_print("")
    #############################################################################################################################






    #############################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def print_list_transforms(self):
        '''
        List all the available data transforms in the selected backend.

        Args:
            None

        Returns:
            None
        '''
        self.custom_print("Transforms List: ");

        if(self.system_dict["library"] == "Mxnet"):
            combined_list_lower = ["apply_random_resized_crop", "apply_center_crop", "apply_color_jitter", "apply_random_horizontal_flip",
                                    "apply_random_vertical_flip", "apply_random_lighting", "apply_resize", "apply_normalize"];

        elif(self.system_dict["library"] == "Keras"):
            combined_list_lower = ["apply_color_jitter", "apply_random_affine", "apply_random_horizontal_flip", 
                                    "apply_random_vertical_flip", "apply_random_rotation", "apply_mean_subtraction", 
                                    "apply_normalize"];


        elif(self.system_dict["library"] == "Pytorch"):
            combined_list_lower = ["apply_center_crop", "apply_color_jitter", "apply_random_affine", "apply_random_crop", 
                                    "apply_random_horizontal_flip", "apply_random_perspective", "apply_random_resized_crop",
                                    "apply_grayscale", "apply_random_rotation", "apply_random_vertical_flip",
                                    "apply_resize", "apply_normalize"];

                                    
        for i in range(len(combined_list_lower)):
            self.custom_print("    {}. {}".format(i+1, combined_list_lower[i]))

        self.custom_print("")
    #############################################################################################################################





    #############################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def print_list_blocks(self):
        '''
        List all the available blocks for custom network creation in the selected backend.

        Args:
            None

        Returns:
            None
        '''
        self.custom_print("Blocks List: ");

        if(self.system_dict["library"] == "Mxnet"):
            combined_list_lower = ["resnet_v1_block", "resnet_v2_block", "resnet_v1_bottleneck_block", "resnet_v2_bottleneck_block",
                                    "resnext_block", "mobilenet_v2_linear_block", "mobilenet_v2_inverted_linear_block",
                                    "squeezenet_fire_block", "conv_bn_relu_block", "inception_a_block", "inception_b_block",
                                    "inception_c_block", "inception_d_block", "inception_e_block"];

        elif(self.system_dict["library"] == "Keras"):
            combined_list_lower = ["resnet_v1_block", "resnet_v2_block", "resnet_v1_bottleneck_block", "resnet_v2_bottleneck_block",
                                    "resnext_block", "mobilenet_v2_linear_block", "mobilenet_v2_inverted_linear_block",
                                    "squeezenet_fire_block", "conv_bn_relu_block", "inception_a_block", "inception_b_block",
                                    "inception_c_block", "inception_d_block", "inception_e_block"];


        elif(self.system_dict["library"] == "Pytorch"):
            combined_list_lower = ["resnet_v1_block", "resnet_v2_block", "resnet_v1_bottleneck_block", "resnet_v2_bottleneck_block",
                                    "resnext_block", "mobilenet_v2_linear_block", "mobilenet_v2_inverted_linear_block",
                                    "squeezenet_fire_block", "conv_bn_relu_block", "inception_a_block", "inception_b_block",
                                    "inception_c_block", "inception_d_block", "inception_e_block"];

                                    
        for i in range(len(combined_list_lower)):
            self.custom_print("    {}. {}".format(i+1, combined_list_lower[i]))

        self.custom_print("")
    #############################################################################################################################


from monk.system.imports import *
from monk.system.base_system_state import update_local_var



#############################################################################################################################
@accepts(str, verbose=[int, str, bool], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def read_json(fname, verbose=0):
    '''
    Read Json file - system dictionary file

    Args:
        fname (str): Path to file containing system states

    Returns:
        dict: loaded system dict
    '''
    with open(fname) as json_file:
        system_dict = json.load(json_file);
        system_dict["verbose"] = verbose;
    return system_dict;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def write_json(system_dict):
    '''
    Write to Json file - system dictionary file

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        None
    '''
    fname = system_dict["fname_relative"];
    f = open(fname, 'w');
    wr = json.dumps(system_dict, indent=4)
    f.write(wr);
    f.close();


@accepts(str, str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def parse_csv(fname, delimiter):
    '''
    Read CSV File - depricted function

    Args:
        fname (str): Path to CSV File
        delimiter (str): Delimiter for csv file

    Returns:
        list: List of images in the csv file
        list: List of corresponding labels
        list: List of all the class names in the dataset
    '''

    #Old format
    #f = open(fname);
    #lst = f.readlines();
    #f.close();
    #del lst[0]
    #img_list = [];
    #label_list = [];
    #for i in range(len(lst)):
    #    img, label = lst[i][:len(lst[i])-1].split(delimiter);
    #    img_list.append(img);
    #    label_list.append(label);
    #classes = list(np.unique(sorted(label_list)))
    #for i in range(len(lst)):
    #    label_list[i] = classes.index(label_list[i]);
    #return img_list, label_list, classes;


    #New format
    df = pd.read_csv(fname);
    columns = df.columns;
    img_list = [];
    label_list = [];
    for i in range(len(df)):
        img_list.append(df[columns[0]][i]);
        label_list.append(str(df[columns[1]][i]));

    classes = list(np.unique(sorted(label_list)))
    if(type(classes[0]) != str):
        classes = list(map(str, classes))
        
    for i in range(len(label_list)):
        label_list[i] = classes.index(label_list[i]);

    return img_list, label_list, classes;



@accepts(str, str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def parse_csv_updated(fname, delimiter):
    '''
    Read CSV File

    Args:
        fname (str): Path to CSV File
        delimiter (str): Delimiter for csv file

    Returns:
        list: List of images in the csv file
        list: List of corresponding labels
        list: List of all the class names in the dataset
    '''
    df = pd.read_csv(fname);
    columns = df.columns;
    img_list = [];
    label_list = [];
    for i in range(len(df)):
        img_list.append(df[columns[0]][i]);
        label_list.append(df[columns[1]][i].split(delimiter));

    classes = [];
    for i in range(len(label_list)):
        tmp = label_list[i];
        for j in range(len(tmp)):
            if(tmp[j] not in classes):
                classes.append(tmp[j]);
    

    return img_list, label_list, sorted(classes);



def parse_csv2(fname, delimiter):
    '''
    Read CSV File - - General Keras

    Args:
        fname (str): Path to CSV File
        delimiter (str): Delimiter for csv file

    Returns:
        df: Dataframe from csv file
        list: List of column names in csv file
    '''
    df = pd.read_csv(fname, delimiter=delimiter);
    df = df.reindex(np.random.permutation(df.index));
    columns = df.columns;
    df[columns[1]] = df[columns[1]].astype(str);
    return df, columns;



def parse_csv2_updated(fname, delimiter):
    '''
    Read CSV File - - Multilabel Keras

    Args:
        fname (str): Path to CSV File
        delimiter (str): Delimiter for csv file

    Returns:
        df: Dataframe from csv file
        list: List of column names in csv file
    '''
    df = pd.read_csv(fname);
    df = df.reindex(np.random.permutation(df.index));
    columns = df.columns;
    df[columns[1]] = df[columns[1]].astype(str);
    df[columns[1]]=df[columns[1]].apply(lambda x:x.split(delimiter))
    return df, columns;



#############################################################################################################################




#############################################################################################################################
@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def save(system_dict):
    '''
    Save system dictionaries

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        None
    '''
    system_dict_copy = system_dict.copy();
    if(system_dict_copy["states"]["eval_infer"]):
        system_dict_tmp = read_json(system_dict_copy["fname_relative"]);
        system_dict_tmp["testing"] = system_dict_copy["testing"];
        system_dict_tmp["dataset"]["test_path"] = system_dict_copy["dataset"]["test_path"];
        system_dict_tmp["dataset"]["transforms"]["test"] = system_dict_copy["dataset"]["transforms"]["test"];
        write_json(system_dict_tmp);  
    else:
        system_dict_copy = update_local_var(system_dict_copy);
        write_json(system_dict_copy);


#############################################################################################################################





#############################################################################################################################
#@accepts(str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=True)
def create_dir(dir_path):
    '''
    Create a new directory

    Args:
        dir_path (str): Directory path and name

    Returns:
        None
    '''
    if(not os.path.isdir(dir_path)):
        os.mkdir(dir_path);  


#@accepts(str, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=True)
def delete_dir(dir_path):
    '''
    Delete an existing directory

    Args:
        dir_path (str): Directory path and name

    Returns:
        None
    '''
    if(os.path.isdir(dir_path)):
        shutil.rmtree(dir_path);

#############################################################################################################################